/**
 ChangeLocationViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import MapKit

class ChangeLocationViewController: BaseViewController {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    let userDefaults = UserDefaults.standard
    
    var searchCompleter = MKLocalSearchCompleter()
    var searchResults = [String]()
    var recentSearchResults = [String]()
    
    var selectedAddress:String?

    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    @IBOutlet weak var searchBar: UISearchBar!
    @IBOutlet weak var tableView: UITableView!
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()

        searchBar.becomeFirstResponder()
        
        if let recentSearch = userDefaults.object(forKey: "RecentSearches") as? [String]{
            recentSearchResults = recentSearch.reversed()
            tableView.reloadData()
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        view.endEditing(true)
    }
    

    // ===========================================================
    // MARK: - Navigation
    // ===========================================================
    
    // In a storyboard-based application, you will often want to do a little preparation before navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if segue is CustomUnwindSegue {
            (segue as! CustomUnwindSegue).animationType = .swipeDown
        }
    }

}


// ===========================================================
// MARK: - UITableView DataSource
// ===========================================================

extension ChangeLocationViewController: UITableViewDataSource {
    
    func numberOfSections(in tableView: UITableView) -> Int {
        if searchResults.count == 0 {
            return 2
        } else {
            return 1
        }
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        
        if searchResults.count == 0 {
            if section == 0 {
                return 1
            } else {
                return recentSearchResults.count
            }
        } else {
            return searchResults.count
        }
        

    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        
        if searchResults.count == 0 {
            if indexPath.section == 0 {
                let cell = UITableViewCell(style: .default, reuseIdentifier: nil)
                cell.textLabel?.font = UIFont(name: "Avenir-Book", size: 15)
                cell.textLabel?.textColor = UIColor.blue
                cell.textLabel?.text = "Current Location"
                // Resize image
                let size = CGSize(width: 16, height: 16)
                cell.imageView?.image = #imageLiteral(resourceName: "CurrentLocation").resizeImage(size: size)
                
                return cell
            } else {
                let cell = UITableViewCell(style: .default, reuseIdentifier: nil)
                cell.textLabel?.font = UIFont(name: "Avenir-Book", size: 15)
                cell.textLabel?.text = recentSearchResults[indexPath.row]
                return cell
            }
        } else {
            let cell = UITableViewCell(style: .default, reuseIdentifier: nil)
            cell.textLabel?.font = UIFont(name: "Avenir-Book", size: 15)
            cell.textLabel?.text = searchResults[indexPath.row]
            return cell
        }
    }
}


// ===========================================================
// MARK: - UITableView Delegate
// ===========================================================

extension ChangeLocationViewController: UITableViewDelegate {
    
    func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        let headerCell = UITableViewCell(style: .default, reuseIdentifier: nil)
        
        headerCell.backgroundColor = UIColor.groupTableViewBackground
        headerCell.textLabel?.font = UIFont(name: "Avenir-Heavy", size: 12)
        headerCell.textLabel?.textColor = UIColor.lightGray
        
        if searchResults.count == 0 {
            headerCell.textLabel?.text = "RECENT SEARCHES"
        } else {
            headerCell.textLabel?.text = "LOCATIONS"
        }
        
        return headerCell
    }
    
    func tableView(_ tableView: UITableView, heightForHeaderInSection section: Int) -> CGFloat {
        
        if searchResults.count == 0 && section == 0 {
            return 0
        } else {
            return 28
        }
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        tableView.deselectRow(at: indexPath, animated: false)
        
        // Manage selected location
        if searchResults.count == 0 {
            if indexPath.section == 0 {
                selectedAddress = "CurrentLocation"
            } else {
                let searchResult = recentSearchResults[indexPath.row]
                selectedAddress = searchResult
            }
        }
        
        // Save recent search if it is necessary
        if searchResults.count != 0 {
            let searchResult = searchResults[indexPath.row]
            if !recentSearchResults.contains(searchResult) {
                recentSearchResults.append(searchResult)
                userDefaults.set(recentSearchResults, forKey: "RecentSearches")
            }
            selectedAddress = searchResult
        }
        
        performSegue(withIdentifier: "unwindToHome", sender: nil)
    }
    
}


// ===========================================================
// MARK: - UISearchBar Delegate
// ===========================================================

extension ChangeLocationViewController: UISearchBarDelegate {

    func searchBar(_ searchBar: UISearchBar, textDidChange searchText: String) {
        
        searchCompleter.delegate = self
        searchCompleter.filterType = .locationsOnly
        searchCompleter.queryFragment = searchBar.text!
    }
}


// ===========================================================
// MARK: - MKLocalSearchCompleter Delegate
// ===========================================================

extension ChangeLocationViewController: MKLocalSearchCompleterDelegate {
    
    func completerDidUpdateResults(_ completer: MKLocalSearchCompleter) {
        searchResults = [String]()
        for searchResult in completer.results {
            if searchResult.subtitle.isEmpty {
             searchResults.append(searchResult.title)
            }
        }
        
        tableView.reloadData()
    }
    
    func completer(_ completer: MKLocalSearchCompleter, didFailWithError error: Error) {
        searchResults = [String]()
        tableView.reloadData()
    }
}

