/**
 EventDetailTableViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import MapKit
import SDWebImage
import MessageUI

class EventDetailTableViewController: UITableViewController {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    // Event selected in list
    var detailEvent: Event?
    
    var delegate: EventDetailProtocol?
    
    var tapRecognizerMap = UITapGestureRecognizer()
    var tapRecognizerCall = UITapGestureRecognizer()
    var tapRecognizerMail = UITapGestureRecognizer()
    var tapRecognizerWebsite = UITapGestureRecognizer()
    
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================

    // Event date label
    @IBOutlet weak var nameLabel: UILabel!
    // Event location label
    @IBOutlet weak var locationShortLabel: UILabel!
    // Event logo image
    @IBOutlet weak var eventLogoImageView: UIImageView!
    // Event logo image
    @IBOutlet weak var eventLogoImageViewBlur: UIImageView!
    // Event date label
    @IBOutlet weak var dateLabel: UILabel!
    // Event location label
    @IBOutlet weak var locationLabel: UILabel!
    
    // Map view to show event location
    @IBOutlet weak var mapView: MKMapView!
    
    // Event type
    @IBOutlet weak var modeIconImageView: UIImageView!
    // Left days label
    @IBOutlet weak var dayLabel: UILabel!
    // Left hours label
    @IBOutlet weak var hourLabel: UILabel!
    // Left minutes label
    @IBOutlet weak var minuteLabel: UILabel!
    
    // Event images collection
    @IBOutlet weak var imagesCollection: UICollectionView!
    
    // Event distance label
    @IBOutlet weak var distanceLabel: UILabel!

    @IBOutlet weak var editionLabel: UILabel!
    @IBOutlet weak var categoryLabel: UILabel!
    @IBOutlet weak var specialityLabel: UILabel!
    @IBOutlet weak var distanceFullLabel: UILabel!
    
    // Call properties
    @IBOutlet weak var callButton: UIView!
    // Mail properties
    @IBOutlet weak var mailButton: UIView!
    // Website properties
    @IBOutlet weak var websiteButton: UIView!
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Init tap gesture recognizer to map
        tapRecognizerMap.addTarget(self, action:#selector(self.showMap))
        mapView.isUserInteractionEnabled = true
        mapView.addGestureRecognizer(tapRecognizerMap)
        
        // Init tap gesture recognizer to call
        tapRecognizerCall.addTarget(self, action:#selector(self.call))
        callButton.isUserInteractionEnabled = true
        callButton.addGestureRecognizer(tapRecognizerCall)
        
        // Init tap gesture recognizer to send an email
        tapRecognizerMail.addTarget(self, action:#selector(self.sendMail))
        mailButton.isUserInteractionEnabled = true
        mailButton.addGestureRecognizer(tapRecognizerMail)
        
        // Init tap gesture recognizer to open a website
        tapRecognizerWebsite.addTarget(self, action:#selector(self.openWebsite))
        websiteButton.isUserInteractionEnabled = true
        websiteButton.addGestureRecognizer(tapRecognizerWebsite)
        
        if detailEvent?.images.count == 0 {
            tableView.deleteSections(IndexSet.init(integer: 1), with: .top)
        }
    }
    
    // Hide Header(s)
    override func tableView(_ tableView: UITableView, heightForHeaderInSection section: Int) -> CGFloat {
        return shouldHideSection(section: section) ? 0.1 : super.tableView(tableView, heightForHeaderInSection: section)
    }
    
    // Hide rows in hidden sections
    override func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return shouldHideSection(section: indexPath.section) ? 0 : super.tableView(tableView, heightForRowAt: indexPath)
    }
    
    // Hide header text by making clear
    override func tableView(_ tableView: UITableView, willDisplayHeaderView view: UIView, forSection section: Int) {
        
        let header = view as! UITableViewHeaderFooterView
        header.textLabel?.font=UIFont(name: "Avenir-Heavy", size: 10)!
        header.textLabel?.textColor=UIColor.gray
        
        if shouldHideSection(section: section) {
            header.textLabel?.textColor = UIColor.clear
        }
    }
    
    /**
     Lanzado por el table view cuando se selecciona una celda
     */
    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        tableView.deselectRow(at: indexPath, animated: false)
    }
    
    // ===========================================================
    // MARK: - OBJC Responds
    // ===========================================================
    
    @objc func showMap() {
        if !detailEvent!.kml.isEmpty {
            delegate?.showMapDetail()
        }
    }
    
    @objc func call() {
        if let error = URL.callPhone(number: detailEvent!.phone) {
            print(error.localizedDescription)
        }
    }
    
    @objc func sendMail() {
        let mailComposeViewController = configuredMailComposeViewController()
        if MFMailComposeViewController.canSendMail() {
            self.present(mailComposeViewController, animated: true, completion: nil)
        } else {
            self.showSendMailErrorAlert()
        }
    }
    
    @objc func openWebsite() {
        
        if let url = URL(string: detailEvent!.website) {
            if let error = url.openBrowser() {
                print(error.localizedDescription)
            }
        }
    }
    

    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    func shouldHideSection(section: Int) -> Bool {
        switch section {
        case 1:  // Hide this section based on condition below
            return detailEvent?.images.count == 0
            
        default:
            return false
        }
    }
    
    func updateEvent(event: Event) {
        detailEvent = event
        updateUI()
    }
    
    /**
     Update UI with event properties
     */
    func updateUI()
    {
        if let event = detailEvent {
            
            // Set event properties
            distanceLabel.text = String(format: "%.0f%@", event.distance, "K")
            nameLabel.text = event.name
            eventLogoImageView.image = event.logo?.image
            eventLogoImageViewBlur.image = event.logo?.image
            dateLabel.text = event.date.dateStringWith(style: .full)
            locationShortLabel.text = event.address?.toString(format: .short)
            locationLabel.text = event.address?.toString(format: .full)
            
            editionLabel.text = String(event.edition)
            distanceFullLabel.text = String(event.distance) + " km"
            categoryLabel.text = event.type?.categoryString
            
            modeIconImageView.image = event.type?.image
            
            setTimeLeft()
            
            Timer.scheduledTimer(timeInterval: 60, target: self, selector: #selector(setTimeLeft), userInfo: nil, repeats: true)
            
            setLocation()
            
            updateBlurEffect()
            
            if event.phone.isEmpty {
                callButton.isUserInteractionEnabled = false
                callButton.alpha = 0.5
            }
            
            if event.mail.isEmpty {
                mailButton.isUserInteractionEnabled = false
                mailButton.alpha = 0.5
            }
            
            if event.website.isEmpty {
                websiteButton.isUserInteractionEnabled = false
                websiteButton.alpha = 0.5
            }
        }
    }
    
    func updateBlurEffect() {
        
        // Get the screen width and height.
        let blurWidth = UIScreen.main.bounds.size.width
        let blurHeight = CGFloat(250)
        
        
        let blur = UIBlurEffect(style: .light)
        let blurView = UIVisualEffectView(effect: blur)
        blurView.frame = CGRect(x: 0, y: 0, width: blurWidth, height: blurHeight)
        eventLogoImageViewBlur.addSubview(blurView)
        eventLogoImageViewBlur.addSubview(blurView)
    }
    
    /**
     Update UI with countdown
     */
    @objc func setTimeLeft() {
        let timeEnd = detailEvent!.date
        let timeNow = Date.init()
        
        
        if timeEnd.compare(timeNow as Date) == ComparisonResult.orderedDescending {
            let calendar = Calendar.current
            let components = calendar.dateComponents([.day, .hour, .minute], from: timeNow, to: timeEnd)
            
            let dayText = String(describing:components.day!)
            let hourText = String(describing:components.hour!)
            let minuteText = String(describing:components.minute!)
            
            dayLabel.text = dayText
            hourLabel.text = hourText
            minuteLabel.text = minuteText
            
            
        } else {
            //timeLeftLabel.text = "Ended"
        }
    }
    
    /**
     Update UI with event location in map view
     */
    func setLocation() {
        //  Set location
        let location = detailEvent!.address!.location.coordinate
        
        let annotation = MKPointAnnotation()
        annotation.coordinate = location
        mapView.addAnnotation(annotation)
        
        let region = MKCoordinateRegion(center: location, span: MKCoordinateSpan(latitudeDelta: 0.1, longitudeDelta: 0.1))
        
        mapView.setRegion(region, animated: true)
    }

}

extension EventDetailTableViewController: MKMapViewDelegate {
    // MARK: - MKMapView
    
    func mapView(_ mapView: MKMapView, viewFor annotation: MKAnnotation) -> MKAnnotationView? {
        
        if let annotationView = mapView.view(for: annotation) {
            annotationView.annotation = annotation
            return annotationView
        } else {
            
            let annotationView = MKAnnotationView(annotation: annotation, reuseIdentifier: nil)
            annotationView.canShowCallout = true
            
            // Resize image
            annotationView.image = #imageLiteral(resourceName: "Location")
            
            return annotationView
        }
    }
}


// ===========================================================
// MARK: - UICollectionView DataSource
// ===========================================================

extension EventDetailTableViewController: UICollectionViewDataSource {
    // MARK: - UICollectionView
    
    /**
     Lanzada por la colección de imágenes cuando necesita saber cuántas debe mostar
     */
    func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
        return detailEvent!.images.count
    }
    
    /**
     Lanzada por la colección de imágenes para que le configuremos cada una de las celdas.
     En nuestro caso cada objeto PlaceImageWithData contiene la imagen completa y una versión thumbnail (reducida).
     En la celda configuraremos la versión thumbnail
     */
    func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        let cell = collectionView.dequeueReusableCell(withReuseIdentifier: "ImageCell", for: indexPath) as! EventImageCell
        
        let eventImage = detailEvent!.images[indexPath.row]
        
        let imageUrl = eventImage.url
        let url = URL(string: imageUrl!)
        
        // Configuramos un indicador de actividad para que se muestre mientras se carga la imagen
        let activityIndicator = UIActivityIndicatorView()
        activityIndicator.center = cell.imageView.center
        activityIndicator.hidesWhenStopped = true
        activityIndicator.color = UIColor.black
        cell.imageView.addSubview(activityIndicator)
        activityIndicator.startAnimating()
        
        // Utilizamos SDWebImage para descargar la imagen en segundo plano directamente sobre el UIImageView de la celda
        cell.imageView.sd_setImage(with: url) {(image: UIImage?, error: Error?, cacheType: SDImageCacheType, imageURL: URL?) -> Void in
            if error != nil {
                print(error!)
            } else {
                //self.eventIma = image!
            }
            // Al terminar la descarga de la imagen quitamos el indicador de actividad
            activityIndicator.removeFromSuperview()
        }
        
        return cell
    }
}


// ===========================================================
// MARK: - UICollectionView Delegate
// ===========================================================

extension EventDetailTableViewController: UICollectionViewDelegate {
    
    /**
     Lanzada por la colección de imágenes cuando se selecciona una de las celdas
     */
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        delegate?.showImages(index: indexPath.row)
    }
}


// ===========================================================
// MARK: - MFMailComposeViewController Delegate
// ===========================================================

extension EventDetailTableViewController: MFMailComposeViewControllerDelegate {
    
    func configuredMailComposeViewController() -> MFMailComposeViewController {
        let mailComposerVC = MFMailComposeViewController()
        mailComposerVC.mailComposeDelegate = self
        
        if let mail = detailEvent?.mail {
            mailComposerVC.setToRecipients([mail])
        }
        mailComposerVC.setSubject("Sending you an in-app e-mail...")
        mailComposerVC.setMessageBody("Sending e-mail in-app is not so bad!", isHTML: false)
        
        return mailComposerVC
    }
    
    func mailComposeController(_ controller: MFMailComposeViewController, didFinishWith result: MFMailComposeResult, error: Error?) {
        controller.dismiss(animated: true, completion: nil)
    }
    
    func showSendMailErrorAlert() {
        self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                            message: NSLocalizedString("ERROR_MAIL", comment: ""),
                            image: nil,
                            buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                            buttonColor: UIColor.primary.normal,
                            style: .default,
                            handler: nil)
    }
}
