/**
 HomeViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */


import UIKit
import Firebase
import GeoFire


class HomeViewController: BaseViewController{
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    // Detail event view
    var detailViewController: EventDetailViewController?
    
    // Search bar controller
    var searchController:UISearchController = UISearchController(searchResultsController: nil)
    
    // Scroll view last Y position
    var lastCheckedContentOffsetY: CGFloat = 0.0
    
    // Events for show in table view
    var events = [Event]()
    
    // Filtered events for show in table view
    var filteredEvents = [Event]()
    
    // Distance between current user location and events location
    var currentDistance = 0.0
    
    // Current user address
    var currentAddress: Address?
    
    // Properties for location service
    var locationManager = CLLocationManager()
    var isUpdatingLocation: Bool = false
    var location: CLLocation?
    var lastLocationError: NSError?
    
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    // Table view
    @IBOutlet weak var tableView: UITableView!
    // Location button
    @IBOutlet weak var locationButton: UIButton!
    // Distance button
    @IBOutlet weak var distanceButton: UIButton!
    
    
    // ===========================================================
    // MARK: - IB Actions
    // ===========================================================
    
    @IBAction func unwind(_ segue: UIStoryboardSegue) {
        if let changeLocation = segue.source as? ChangeLocationViewController,
            let selectedAddress = changeLocation.selectedAddress {
            events = [Event]()
            updateTable()
            if selectedAddress == "CurrentLocation" {
                checkLocationAuthorizationStatus()
            } else {
                queryAddress(address: selectedAddress)
            }
        }
    }
    
    /**
     Search button action
     */
    @IBAction func actionSearch(_ sender: AnyObject) {
        showSearchBar()
    }
    
    /**
     Location button action
     */
    @IBAction func actionChangeLocation(_ sender: AnyObject) {
        self.performSegue(withIdentifier:"changeLocation", sender: nil)
    }
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Configure location button
        locationButton.cornerRadius = locationButton.bounds.height/2
        locationButton.setTitle("  Current Location  \u{25BE}  ", for: .normal)
        locationButton.sizeToFit()
        
        // Configure floating distance button
        distanceButton.setShadowButton()
        distanceButton.isHidden = true
        
        // Configure split view
        if let splitViewController = splitViewController {
            let controllers = splitViewController.viewControllers
            detailViewController = (controllers[controllers.count-1] as! UINavigationController).topViewController as? EventDetailViewController
        }
        
        // Start to search current location and update UI
        checkLocationAuthorizationStatus()
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        navigationController?.hidesBarsOnSwipe = true
        self.view.solidStatusBar(backgroundColor: UIColor.white)
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        tabBarController?.tabBar.isHidden = false
        if let row = tableView.indexPathForSelectedRow {
            self.tableView.deselectRow(at: row, animated: false)
        }
    }
    
    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    /**
     Update table with new values.
    */
    func updateTable() {
        tableView.reloadData()
    }
    
    /**
     Update distance label, depends of first event showed.
     
     - parameter distance: Distance to be updated.
    */
    func updateDistance(distance: Double) {
        
        let roundedDistance = distance.roundDownExp()
        
        currentDistance = roundedDistance
        
        distanceButton.setTitle(String(format: "  %.0f+ %@  ", roundedDistance, "km from you"), for: .normal)
        distanceButton.sizeToFit()
    }
    
    /**
     Update current location label.
     
     - parameter error: Error to reverse it.
     */
    func updateLocation(error: Error?) {
        if let error = error {
            print(error.localizedDescription)
            return
        }
        if let current = currentAddress {
            let title = String(format: "  %@  \u{25BE}  ", current.toString(format: .short))
            locationButton.setTitle(title, for: .normal)
            locationButton.sizeToFit()
            
            queryEvents(location: current.location)
        }
    }
    
    /**
     Show search bar.
    */
    func showSearchBar() {
        
        // Configure search bar
        searchController.hidesNavigationBarDuringPresentation = true
        searchController.searchBar.barTintColor = navigationController?.navigationBar.barTintColor
        searchController.searchBar.tintColor = navigationController?.navigationBar.tintColor
        searchController.searchBar.delegate = self
        searchController.searchResultsUpdater = self
        
        present(searchController, animated: true, completion: nil)
    }
    
    /**
     Filter the initial event list.
     
     - parameter searchText: Text to be check.
     - parameter scope: Type of event.
     */
    func filterContentForSearchText(searchText: String, scope: String = "All") {
        filteredEvents = events.filter({( event : Event) -> Bool in
            let categoryMatch = (scope == "All") //|| (event.category == scope)
            return categoryMatch && event.name.lowercased().contains(searchText.lowercased())
        })
        updateTable()
    }
    
    
    // ===========================================================
    // MARK: - Queries
    // ===========================================================
    
    /**
     Query all events that are near to the current location (geofire service).
     */
    func queryEvents(location: CLLocation) {
        // Indicator showed to user
        activityIndicator.show(text: "Loading Events")
        
        FirebaseHelper.sharedInstance.getAllEvents(center: location) { event in
            
            self.events.append(event)
            
            // Be sure that the tableView updates when there is new data.
            self.updateTable()
            self.activityIndicator.hide()
        }
    }
    
    /**
     Query information about current location.
    */
    func queryLocation(location: CLLocation) {
        activityIndicator.show(text: "Update Location")
        currentAddress = Address(location: location) { error in
            self.updateLocation(error: error)
        }
    }
    
    /**
     Query information about current location.
     */
    func queryAddress(address: String) {
        activityIndicator.show(text: "Update Location")
        currentAddress = Address(address: address) { error in
            self.updateLocation(error: error)
        }
    }
    
    // ===========================================================
    // MARK: - Navigation
    // ===========================================================

    /**
     Go to event detail view. 
     
     - parameter event: Information about selected event.
    */
    func showDetailEvent(event: Event) {
        self.performSegue(withIdentifier:"showDetail", sender: event)
    }
    
    
    // In a storyboard-based application, you will often want to do a little preparation before navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if (segue.identifier == "showDetail") {
            if let controller = (segue.destination as! UINavigationController).topViewController as? EventDetailViewController, let event = sender as? Event {
                controller.detailEvent = event
                controller.navigationItem.leftBarButtonItem = splitViewController?.displayModeButtonItem
                controller.navigationItem.leftItemsSupplementBackButton = true
            }
        }
        
        if segue is CustomStoryboardSegue {
            (segue as! CustomStoryboardSegue).animationType = .swipeDown
        }
    }
}


// ===========================================================
// MARK: - UITableView DataSource
// ===========================================================

extension HomeViewController: UITableViewDataSource {
    
    func numberOfSections(in tableView: UITableView) -> Int {
        if searchController.isActive && searchController.searchBar.text != "" {
            return filteredEvents.count
        }
        
        return events.count
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return 1
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "eventCell", for: indexPath) as! EventCell
        
        cell.currentLocation = currentAddress?.location
        
        if searchController.isActive && searchController.searchBar.text != "" {
            cell.event = filteredEvents[indexPath.section]
        } else {
            cell.event = events[indexPath.section]
        }
        
        return cell
    }
}


// ===========================================================
// MARK: - UITableView Delegate
// ===========================================================

extension HomeViewController: UITableViewDelegate {
    
    func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        let  headerCell = tableView.dequeueReusableCell(withIdentifier: "eventHeaderCell") as! EventHeaderCell
        
        if searchController.isActive && searchController.searchBar.text != "" {
            headerCell.event  = filteredEvents[section]
        } else {
            headerCell.event = events[section]
        }
        
        return headerCell
    }
    
    func tableView(_ tableView: UITableView, heightForHeaderInSection section: Int) -> CGFloat {
        return 60
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        tableView.deselectRow(at: indexPath, animated: false)
        let event = events[indexPath.section]
        showDetailEvent(event: event)
    }
}


// ===========================================================
// MARK: - UIScrollView Delegate
// ===========================================================

extension HomeViewController {
    
    /**
     Triggered when a user move scroll.
     */
    func scrollViewDidScroll(_ scrollView: UIScrollView) {
        
        // Save new Y
        let newContentOffsetY = tableView.contentOffset.y
        
        if newContentOffsetY <= -50 {
            // Scroll at first row, hide distance
            distanceButton.isHidden = true
            return
        }
        
        
        // Check scroll up
        let tableViewHasScrollUpFarEnough = newContentOffsetY > lastCheckedContentOffsetY + 50
        
        // Check scroll down
        let tableViewHasScrollDownFarEnough = tableView.contentOffset.y < lastCheckedContentOffsetY - 50
        
        // If user scroll is +- 50 difference to last saved Y
        let tableViewHasScrollFarEnough = tableViewHasScrollUpFarEnough || tableViewHasScrollDownFarEnough
        
        if (!tableViewHasScrollFarEnough) {
            // Scroll is not far enough
            return
        }
        
        
        // Save Y to last checked
        self.lastCheckedContentOffsetY = tableView.contentOffset.y
        
        
        // Get visible cells
        let visibleCells = tableView.visibleCells as! [EventCell]
        
        if let cell = visibleCells.first {
            
            var distance = 0.0
            
            if visibleCells.count > 2 {
                // More than 2 cells on view
                
                //Calculate first cell visibility.
                let convertedRect = tableView.convert(cell.frame, to: tableView.superview)
                let visibleHeight = tableView.frame.intersection(convertedRect).height
                
                if visibleHeight > cell.frame.height*0.3 {
                    distance = cell.distanceCurrentLocation
                } else {
                    distance = visibleCells[1].distanceCurrentLocation
                }
            } else {
                // 1 or 2 cells on view
                distance = cell.distanceCurrentLocation
            }
            
            // Format result distance
            distance = distance.roundDownExp()
            
            distanceButton.isHidden = distance <= 0
            
            if distance != currentDistance {
                updateDistance(distance: distance)
            }
        }
    }
}


// ===========================================================
// MARK: - UISearchResults Updating
// ===========================================================

extension HomeViewController: UISearchResultsUpdating{
    
    func updateSearchResults(for searchController: UISearchController) {
        filterContentForSearchText(searchText: searchController.searchBar.text!)
    }
}


// ===========================================================
// MARK: - UISearchBar Delegate
// ===========================================================

extension HomeViewController: UISearchBarDelegate {
    
    func searchBar(_ searchBar: UISearchBar, selectedScopeButtonIndexDidChange selectedScope: Int) {
        filterContentForSearchText(searchText: searchBar.text!, scope: searchBar.scopeButtonTitles![selectedScope])
    }
}


// ===========================================================
// MARK: - CLLocationManager Delegate
// ===========================================================

extension HomeViewController: CLLocationManagerDelegate {
    
    func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {
        if status == .authorizedAlways || status == .authorizedWhenInUse {
            location = nil
            startLocationManager()
        }
    }
    
    /**
     It is called every time that gets new user location.
     If everything is ok, it get a better location each time, when we have the desired location, we will
     stop all the process.
     */
    func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]){
        // Update the display with the new location data
        let newLocation = locations.last!
        print("didUpdateLocations \(newLocation)")
        
        // Test the age of the location measurement to determine if the measurement is cached.
        // Accept locations received within the last 5 minutes.
        if newLocation.timestamp.timeIntervalSinceNow < -300 {
            return
        }
        
        // Test that the horizontal accuracy does not indicate an invalid measurement
        if newLocation.horizontalAccuracy < 0 {
            return
        }
        
        // Test the measurement to see if it is more accurate than the previous measurement
        if location == nil || location!.horizontalAccuracy > newLocation.horizontalAccuracy {
            lastLocationError = nil
            location = newLocation
        }
        
        // Test the measurement to see if it meets the desired accuracy
        if newLocation.horizontalAccuracy <= kCLLocationAccuracyThreeKilometers {
            print("Got the desired accuracy")
            
            // We have a measurement that meets our requirements, so we can stop updating the location
            stopLocationManager()
            
            if let location = location {
                queryLocation(location: location)
            } else {
                activityIndicator.hide()
            }
        }
    }
    
    /**
     It is called when location manager fire a error trigger.
     */
    func locationManager(_ manager: CLLocationManager, didFailWithError error: Error) {
        
        let nsError = error as NSError
        if nsError.code == CLError.locationUnknown.rawValue {
            // No location found it, location manager will try to search again.
            return
        }
        
        // Geolocation error, stop the process and notify the user.
        lastLocationError = nsError
        stopLocationManager()
        
        self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                            message: error.localizedDescription,
                            image: nil,
                            buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                            buttonColor: UIColor.primary.normal,
                            style: .default,
                            handler: nil)
        
        //showErrorAlert(title: "Geolocation error", titleButton: "OK", message: "It's not possible to obtain your current location", style: .default, handler: nil)
    }
    
    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    private func startLocationManager() {
        if CLLocationManager.locationServicesEnabled() {
            locationManager.desiredAccuracy = kCLLocationAccuracyKilometer
            locationManager.startUpdatingLocation()
            isUpdatingLocation = true
        }
    }
    
    /**
     Stop process to get user location.
     */
    private func stopLocationManager() {
        if isUpdatingLocation {
            locationManager.stopUpdatingLocation()
            isUpdatingLocation = false
        }
    }
    
    /**
     Init search current location.
     Check authoritzation status and then start location manager.
     */
    fileprivate func checkLocationAuthorizationStatus() {
        
        activityIndicator.show(text: "Update Location")
        locationManager.delegate = self
        
        // Ask user authoritzation (iOS 7+)
        let authStatus = CLLocationManager.authorizationStatus()
        if authStatus == .notDetermined {
            locationManager.requestWhenInUseAuthorization()
            return
        }
        if authStatus == .denied || authStatus == .restricted {
            
            self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                message: "error.localizedDescription",
                                image: nil,
                                buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                buttonColor: UIColor.primary.normal,
                                style: .default,
                                handler: nil)
            //showErrorAlert(title: "Location Services Disabled", titleButton: "OK", message: "Please enable location services for this app in Settings.", style: .default, handler: nil)
            return
        }
        if authStatus == .authorizedAlways || authStatus == .authorizedWhenInUse {
            location = nil
            startLocationManager()
        }
    }
}

