/**
 MapViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import MapKit

class MapViewController: BaseViewController{
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    // Event selected in list
    var detailEvent: Event?
    
    fileprivate var kmlParser: KMLParser!
    
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    // Map view to show event location
    @IBOutlet weak var mapView: MKMapView!

    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        //  Set location
        let location = detailEvent!.address!.location.coordinate
        
        let region = MKCoordinateRegion(center: location, span: MKCoordinateSpan(latitudeDelta: 0.1, longitudeDelta: 0.1))
        
        mapView.setRegion(region, animated: true)
        
        DispatchQueue.global(qos: .default).async {
            // Locate the path to the   file in the application's bundle
            // and parse it with the KMLParser.
            let url = URL(string: self.detailEvent!.kml)
            self.kmlParser = KMLParser(URL: url!)
            self.kmlParser.parseKML()
            
            DispatchQueue.main.async {
                // Add all of the MKOverlay objects parsed from the KML file to the map.
                let overlays = self.kmlParser.overlays
                self.mapView.addOverlays(overlays)
                
                // Add all of the MKAnnotation objects parsed from the KML file to the map.
                let annotations = self.kmlParser.points
                self.mapView.addAnnotations(annotations)
                
                // Walk the list of overlays and annotations and create a MKMapRect that
                // bounds all of them and store it into flyTo.
                var flyTo = MKMapRectNull
                for overlay in overlays {
                    if MKMapRectIsNull(flyTo) {
                        flyTo = overlay.boundingMapRect
                    } else {
                        flyTo = MKMapRectUnion(flyTo, overlay.boundingMapRect)
                    }
                }
                
                for annotation in annotations {
                    let annotationPoint = MKMapPointForCoordinate(annotation.coordinate)
                    let pointRect = MKMapRectMake(annotationPoint.x, annotationPoint.y, 0, 0)
                    if MKMapRectIsNull(flyTo) {
                        flyTo = pointRect
                    } else {
                        flyTo = MKMapRectUnion(flyTo, pointRect)
                    }
                }
                
                // Position the map so that all overlays and annotations are visible on screen.
                self.mapView.visibleMapRect = flyTo
            }
        }
    }
}


// ===========================================================
// MARK: - MKMapView Delegate
// ===========================================================

extension MapViewController: MKMapViewDelegate {
    // MARK: - MKMapView
    
    func mapView(_ mapView: MKMapView, rendererFor overlay: MKOverlay) -> MKOverlayRenderer {
        return kmlParser.rendererForOverlay(overlay)!
    }
    
    func mapView(_ mapView: MKMapView, viewFor annotation: MKAnnotation) -> MKAnnotationView? {
        return kmlParser.viewForAnnotation(annotation)
    }
}
