/**
 LaunchViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import NVActivityIndicatorView
import Firebase

class LaunchViewController: BaseViewController {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    let maxSize: CGFloat = 200
    
    private let mask = CALayer()
    
    fileprivate var idSegue = ""
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        
        // Control iPhone/iPad
        var size:CGFloat = self.view.frame.size.width*0.3
        if (size > maxSize) {
            size = maxSize
        }
        
        activityIndicatorSize = size
        
        super.viewDidLoad()
        
        ///////////////////////////////
        // CONFIGURATION MASK OF LOGO
        
        // Create mask bounds
        let logoBounds = CGRect(x: 0, y: 0, width: size, height: size)
        
        // Configure mask
        mask.contents = #imageLiteral(resourceName: "Logo").cgImage
        mask.bounds = logoBounds
        mask.anchorPoint = CGPoint(x: 0.5, y: 0.5)
        mask.position = self.view.center
        
        // Add mask to view
        let logoMask = UIView(frame: self.view.frame)
        logoMask.backgroundColor = #colorLiteral(red: 1.0, green: 1.0, blue: 1.0, alpha: 1.0)
        logoMask.layer.mask = mask
        
        // Add logo mask to parent view
        self.view.addSubview(logoMask)
        
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        
        if (UIApplication.isFirstLaunch()) { // Go to welcome
            self.showWelcomeSlides()
        } else if (Auth.auth().currentUser == nil) { // Go to sign in
            self.showSignIn()
        } else { // Go to main
            self.showMain()
        }
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    
    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    /**
     Animate logo transition between storyboards.
     */
    func startLogoAnimation() {
        
        ///////////////////////////////
        // Create bounds animation, it is like Twitter app
        
        let boundsAnimation = CAKeyframeAnimation(keyPath: "bounds")
        boundsAnimation.duration = 1.5
        boundsAnimation.timingFunctions = [CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut), CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut)]
        let maxSize = max(self.view.frame.size.width, self.view.frame.size.height)
        boundsAnimation.values = [NSValue(cgRect: mask.bounds),
                                  NSValue(cgRect: CGRect(x: 0, y: 0, width: mask.bounds.width/3, height: mask.bounds.height/3)),
                                  NSValue(cgRect: CGRect(x: 0, y: 0, width: maxSize+1000, height: maxSize+1000))]
        boundsAnimation.keyTimes = [0, 0.3, 1]
        boundsAnimation.fillMode = kCAFillModeForwards
        boundsAnimation.isRemovedOnCompletion = false
        
        ///////////////////////////////
        // Create opacity animation, it is to generate difumination effect
        
        let opacityAnimation = CAKeyframeAnimation(keyPath: "opacity")
        opacityAnimation.duration = 1
        opacityAnimation.timingFunctions = [CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut), CAMediaTimingFunction(name: kCAMediaTimingFunctionEaseInEaseOut)]
        opacityAnimation.values = [1, 1, 0]
        opacityAnimation.keyTimes = [0, 0.3, 1]
        opacityAnimation.fillMode = kCAFillModeForwards
        opacityAnimation.isRemovedOnCompletion = false
        opacityAnimation.delegate = self
        
        // Add animations to current mask
        mask.add(boundsAnimation, forKey: "bounds")
        mask.add(opacityAnimation, forKey: "opacity")
    }
    
    
    // ===========================================================
    // MARK: - Navigation
    // ===========================================================
    
    /**
     Go to welcome slides views.
     */
    func showWelcomeSlides() {
        idSegue = "showWelcomeSlides"
        activityIndicator.hide()
        startLogoAnimation()
    }
    
    /**
     Go to sign in view.
     */
    func showSignIn() {
        idSegue = "showSignIn"
        activityIndicator.hide()
        startLogoAnimation()
    }
    
    /**
     Go to main.
     */
    func showMain() {
        idSegue = "showMain"
        activityIndicator.hide()
        startLogoAnimation()
    }
}


// ===========================================================
// MARK: - CAAnimation Delegate
// ===========================================================

extension LaunchViewController: CAAnimationDelegate {
    
    func animationDidStop(_ anim: CAAnimation, finished flag: Bool) {
        
        // If animation is finished, perform segue
        if (flag) {
            self.performSegue(withIdentifier: idSegue, sender: nil)
        }
    }
}



