/**
 WebViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import WebKit
import NVActivityIndicatorView

enum WebMode {
    case termsConditions
    case privacyPolicy
    case faq
}

class WebViewController: BaseViewController {
    
    // ===========================================================
    // Variables and constants
    // ===========================================================
    
    // Terms and conditions mode
    var mode: WebMode = .termsConditions
    
    var fullURL = ""
    
    var webView: WKWebView?
    
    // ===========================================================
    // UI Outlets
    // ===========================================================
    
    
    // ===========================================================
    // UI Actions
    // ===========================================================
    
    @objc func back(sender: UIBarButtonItem) {
        self.performSegue(withIdentifier: "unwindAny", sender: nil)
    }
    
    // ===========================================================
    // UIView Lifecycle
    // ===========================================================
    
    override func loadView() {
        let webConfiguration = WKWebViewConfiguration()
        webView = WKWebView(frame: .zero, configuration: webConfiguration)
        webView?.uiDelegate = self
        webView?.navigationDelegate = self
        view = webView
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        self.navigationItem.hidesBackButton = true
        let newBackButton = UIBarButtonItem(title: "", style: UIBarButtonItemStyle.plain, target: self, action: #selector(back(sender:)))
        newBackButton.tintColor = #colorLiteral(red: 1.0, green: 1.0, blue: 1.0, alpha: 1.0)
        newBackButton.image = #imageLiteral(resourceName: "Arrow_Down")
        self.navigationItem.leftBarButtonItem = newBackButton
        
        switch mode {
        case .termsConditions:
            self.title = NSLocalizedString("TERMS_SERVICE", comment: "")
            fullURL = "https://www.google.es"
        case .privacyPolicy:
            self.title = NSLocalizedString("PRIVACY_POLICY", comment: "")
            fullURL = "https://www.google.es"
        case .faq:
            self.title = "FAQ"
            fullURL = "https://www.google.es"
        }
        
        // Clear cache if there ara network connection and 48h from now
        
        var lastCleaningTimestamp = Date(mysqlTimestamp: "1970-01-01 00:00:01")
        
        if let timestamp = UserDefaults.standard.object(forKey: Constants.cache) as? Date {
            lastCleaningTimestamp = timestamp
        }
        
        if ( Date().hours(from: lastCleaningTimestamp) > 48 && currentReachabilityStatus != .notReachable) {
            clearCache()
            // Update local timestamp
            UserDefaults.standard.set(Date(), forKey: Constants.cache)
        }
        
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        
        let url = URL(string: fullURL)
        let request = URLRequest(url: url!)
        
        //Load web
        webView?.load(request)
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        // Hide the navigation bar on the this view controller
        self.navigationController?.setNavigationBarHidden(false, animated: animated)
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        // Show the navigation bar on other view controllers
        self.navigationController?.setNavigationBarHidden(true, animated: animated)
    }
    
    // ===========================================================
    // Methods
    // ===========================================================
    
    func clearCache() {
        let websiteDataTypes = NSSet(array: [WKWebsiteDataTypeDiskCache, WKWebsiteDataTypeMemoryCache])
        let date = NSDate(timeIntervalSince1970: 0)
        
        WKWebsiteDataStore.default().removeData(ofTypes: websiteDataTypes as! Set<String>, modifiedSince: date as Date, completionHandler:{ })
    }
    
    
    // ===========================================================
    // MARK: - Navigation
    // ===========================================================
    
    // In a storyboard-based application, you will often want to do a little preparation before navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if segue is CustomUnwindSegue {
            (segue as! CustomUnwindSegue).animationType = .swipeDown
        }
    }
}

extension WebViewController : WKUIDelegate, WKNavigationDelegate {
    
    func webView(_ webView: WKWebView, didStartProvisionalNavigation navigation: WKNavigation!) {
        activityIndicator.show(text: "")
    }
    
    func webView(_ webView: WKWebView, didFinish navigation: WKNavigation!) {
        activityIndicator.hide()
    }
    
    func webView(_ webView: WKWebView, didFail navigation: WKNavigation!, withError error: Error) {
        activityIndicator.hide()
        self.showNotificationAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                   message: NSLocalizedString("BAD_NETWORK", comment: ""),
                                   theme: .warning)
    }
    
    func webView(_ webView: WKWebView, decidePolicyFor navigationAction: WKNavigationAction, decisionHandler: @escaping (WKNavigationActionPolicy) -> Void){
        print("webView:\(webView) decidePolicyForNavigationAction:\(navigationAction) decisionHandler:\(decisionHandler)")
        
        switch navigationAction.navigationType {
        case .linkActivated:
            if navigationAction.targetFrame == nil {
                webView.load(navigationAction.request)
            }
            if let url = navigationAction.request.url {
                _ = url.openBrowser()
                decisionHandler(.cancel)
                return
            }
        default:
            break
        }
        
        if let url = navigationAction.request.url {
            print(url.absoluteString)
            
        }
        decisionHandler(.allow)
    }
    
}

