/**
 ResetPasswordViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import Firebase
import NVActivityIndicatorView

class ResetPasswordViewController: BaseViewController {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    //  Email passed by segue
    var email: String!
    
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    //  Email field
    @IBOutlet weak var fieldEmail: AMInputView!
    //  Reset action button
    @IBOutlet weak var buttonReset: UIButton!
    
    
    // ===========================================================
    // MARK: - IB Actions
    // ===========================================================
    
    /**
     Reset button action.
     */
    @IBAction func actionReset(_ sender: AnyObject) {
        resetPassword()
    }
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        // Add new functionality to view.
        self.view.hideKeyboardWhenTappedAround()
        
        // Set delegate
        fieldEmail.delegate = self
        
        // Add text passed by segue
        fieldEmail.textFieldView.text = email
        
        // Apply expand mode
        fieldEmail.setViewExpandMode(expand: fieldEmail.textFieldView.hasText)
        
        // Block reset button.
        buttonReset.isEnabled = fieldEmail.textFieldView.hasText
        
        let swipeUp = UISwipeGestureRecognizer(target: self, action: #selector(self.respondToSwipeGesture))
        swipeUp.direction = .up
        self.view.addGestureRecognizer(swipeUp)
        
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        view.endEditing(true)
    }
    
    
    // ===========================================================
    // MARK: - OBJC Responds
    // ===========================================================
    
    @objc func respondToSwipeGesture(gesture: UIGestureRecognizer) {
        if let swipeGesture = gesture as? UISwipeGestureRecognizer {
            switch swipeGesture.direction {
            case .up:
                self.performSegue(withIdentifier: "unwindResetPassword", sender: nil)
            default:
                break
            }
        }
    }
    
    
    // ===========================================================
    // MARK: - Firebase
    // ===========================================================
    
    /**
     Check email entered for the user and then try requests a "password reset" email be sent
     */
    func resetPassword() {
        let email = fieldEmail.textFieldView.text!
        
        // Array of errors detected in fields
        var errors = [String]()
        
        // Paint colors for error results
        let colorDefault = UIColor.clear
        let colorError = UIColor.red.withAlphaComponent(0.40)
        
        // Check email field
        if email.isEmpty {
            fieldEmail.backgroundColor = colorError
            errors.append("The email can't be empty")
        } else if !email.isEmailValid {
            fieldEmail.backgroundColor = colorError
            errors.append("The email have invalid sintax")
        } else {
            fieldEmail.backgroundColor = colorDefault
        }
        
        let numErrors = errors.count
        
        if numErrors > 0 {
            // Inform users about field errors
            showPopupAlert(title: NSLocalizedString("REVIEW_FORM", comment: ""),
                           message: errors.joined(separator: ".\n"),
                           image: nil,
                           buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                           buttonColor: UIColor.primary.normal,
                           style: .default,
                           handler: nil)
        } else {
            
            UIApplication.shared.beginIgnoringInteractionEvents()
            activityIndicator.show(text: "")
            
            Auth.auth().sendPasswordReset(withEmail: email) { (error) in
                UIApplication.shared.endIgnoringInteractionEvents()
                self.activityIndicator.hide()
                
                if let error = error {
                    self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                        message: error.localizedDescription,
                                        image: nil,
                                        buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                        buttonColor: UIColor.primary.normal,
                                        style: .default,
                                        handler: nil)
                } else {
                    self.showPopupAlert(title: NSLocalizedString("EMAIL_RESET_TITLE", comment: ""),
                                        message: NSLocalizedString("EMAIL_RESET_SUBTITLE", comment: ""),
                                        image: nil,
                                        buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                        buttonColor: UIColor.primary.normal,
                                        style: .default) {
                                            
                                            self.performSegue(withIdentifier: "unwindResetPassword", sender: nil)
                    }
                }
            }
        }
    }
    
    
    // ===========================================================
    // MARK: - Navigation
    // ===========================================================
    
    // In a storyboard-based application, you will often want to do a little preparation before navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if segue is CustomUnwindSegue {
            (segue as! CustomUnwindSegue).animationType = .swipeUp
        }
    }
}


// ===========================================================
// MARK: - AMInput Delegate
// ===========================================================

extension ResetPasswordViewController: AMInputDelegate {
    
    func amInputView(hasText: Bool) {
        // Block reset button.
        buttonReset.isEnabled = hasText
    }
}
