/**
 SignInViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import Firebase
import FBSDKLoginKit
import GoogleSignIn
import NVActivityIndicatorView

enum AMSigninSignupViewMode {
    case signin
    case signup
}

class SignInViewController: BaseViewController{
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    let animationDuration = 0.25
    var mode:AMSigninSignupViewMode = .signup
    
    // Image tap gesture recognizer
    var tapRecognizerFacebook = UITapGestureRecognizer()
    var tapRecognizerGoogle = UITapGestureRecognizer()
    var screenEdgeRecognizer = UIScreenEdgePanGestureRecognizer()
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    // Background image constraints
    @IBOutlet weak var backImageLeftConstraint: NSLayoutConstraint!
    @IBOutlet weak var backImageBottomConstraint: NSLayoutConstraint!
    
    // Login views and constrains
    @IBOutlet weak var signinView: UIView!
    @IBOutlet weak var signinContentView: UIView!
    @IBOutlet weak var signinButton: UIButton!
    @IBOutlet weak var emailField: AMInputView!
    @IBOutlet weak var passwordField: AMInputView!
    @IBOutlet weak var signinButtonVerticalCenterConstraint: NSLayoutConstraint!
    @IBOutlet weak var signinButtonTopConstraint: NSLayoutConstraint!
    @IBOutlet weak var signinWidthConstraint: NSLayoutConstraint!
    
    // Signup views and constrains
    @IBOutlet weak var signupView: UIView!
    @IBOutlet weak var signupContentView: UIView!
    @IBOutlet weak var signupButton: UIButton!
    @IBOutlet weak var emailFieldS: AMInputView!
    @IBOutlet weak var passwordFieldS: AMInputView!
    @IBOutlet weak var signupButtonVerticalCenterConstraint: NSLayoutConstraint!
    @IBOutlet weak var signupButtonTopConstraint: NSLayoutConstraint!
    
    // Logo and constrains
    @IBOutlet weak var logoView: UIView!
    @IBOutlet weak var logoTopConstraint: NSLayoutConstraint!
    @IBOutlet weak var logoHeightConstraint: NSLayoutConstraint!
    @IBOutlet weak var logoBottomConstraint: NSLayoutConstraint!
    @IBOutlet weak var logoButtomInSingupConstraint: NSLayoutConstraint!
    @IBOutlet weak var logoCenterConstraint: NSLayoutConstraint!
    
    //  General and constrains
    @IBOutlet weak var forgotPassTopConstraint: NSLayoutConstraint!
    @IBOutlet weak var socialsView: UIView!
    @IBOutlet weak var facebookBotton: UIImageView!
    @IBOutlet weak var googleButton: UIImageView!
    
    
    // ===========================================================
    // MARK: - IB Actions
    // ===========================================================
    
    @IBAction func signinAction(_ sender: Any) {
        
        if mode == .signup {
            toggleViewMode(animated: true)
        } else {
            signin()
        }
    }
    
    @IBAction func signupAction(_ sender: Any) {
        
        if mode == .signin {
            toggleViewMode(animated: true)
        } else {
            signup()
        }
    }
    
    @IBAction func TermsServiceAction(_ sender: Any) {
        
        self.performSegue(withIdentifier: "showWeb", sender: WebMode.termsConditions)
    }
    
    @IBAction func PrivacyPolicyAction(_ sender: Any) {
        
        self.performSegue(withIdentifier: "showWeb", sender: WebMode.privacyPolicy)
    }
    
    @IBAction func unwind(segue: UIStoryboardSegue) {
        
    }
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        //  Add new functionality to view.
        self.view.hideKeyboardWhenTappedAround()
        
        // Set view to login mode
        toggleViewMode(animated: false)
        
        // Add keyboard notification
        NotificationCenter.default.addObserver(self, selector: #selector(keyboarFrameChange(notification:)), name: .UIKeyboardWillChangeFrame, object: nil)
        
        // Configure Google delegates
        GIDSignIn.sharedInstance().delegate = self
        GIDSignIn.sharedInstance().uiDelegate = self
        
        // Init tap gesture recognizer to select login Facebook
        tapRecognizerFacebook.addTarget(self, action:#selector(self.loginFacebook))
        facebookBotton.isUserInteractionEnabled = true
        facebookBotton.addGestureRecognizer(tapRecognizerFacebook)
        
        // Init tap gesture recognizer to select login Google
        tapRecognizerGoogle.addTarget(self, action:#selector(self.loginGoogle))
        googleButton.isUserInteractionEnabled = true
        googleButton.addGestureRecognizer(tapRecognizerGoogle)
        
        // Init screen edge recognizer
        screenEdgeRecognizer = UIScreenEdgePanGestureRecognizer(target: self, action:#selector(self.togleFromSwipe))
        screenEdgeRecognizer.edges = .right
        view.addGestureRecognizer(screenEdgeRecognizer)
    
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        view.endEditing(true)
    }
    
    // ===========================================================
    // MARK: - Toggle view
    // ===========================================================
    
    @objc func togleFromSwipe(sender: UIScreenEdgePanGestureRecognizer) {
        if sender.state == .ended {
            toggleViewMode(animated: true)
        }
    }
    
    func toggleViewMode(animated:Bool){
        
        // toggle mode
        if mode == .signin {
            mode = .signup
            screenEdgeRecognizer.edges = .left
        } else {
            mode = .signin
            screenEdgeRecognizer.edges = .right
        }
        
        
        // set constraints changes
        backImageLeftConstraint.constant = mode == .signin ? 0:-self.view.frame.size.width
        
        
        signinWidthConstraint.isActive = mode == .signup ? true:false
        logoCenterConstraint.constant = (mode == .signin ? -1:1) * (signinWidthConstraint.multiplier * self.view.frame.size.width)/2
        signinButtonVerticalCenterConstraint.priority = mode == .signin ? UILayoutPriority(300):UILayoutPriority(900)
        signupButtonVerticalCenterConstraint.priority = mode == .signup ? UILayoutPriority(300):UILayoutPriority(900)
        
        
        //animate
        self.view.endEditing(true)
        
        UIView.animate(withDuration:animated ? animationDuration:0) {
            
            //animate constraints
            self.view.layoutIfNeeded()
            
            //hide or show views
            self.signinContentView.alpha = self.mode == .signin ? 1:0
            self.signupContentView.alpha = self.mode == .signup ? 1:0
            
            
            // rotate and scale login button
            let scaleLogin:CGFloat = self.mode == .signin ? 1:0.5
            let rotateAngleLogin:CGFloat = self.mode == .signin ? 0: -(CGFloat.pi / 2)
            
            var transformLogin = CGAffineTransform(scaleX: scaleLogin, y: scaleLogin)
            transformLogin = transformLogin.rotated(by: rotateAngleLogin)
            self.signinButton.transform = transformLogin
            
            
            // rotate and scale signup button
            let scaleSignup:CGFloat = self.mode == .signup ? 1:0.5
            let rotateAngleSignup:CGFloat = self.mode == .signup ? 0: -(CGFloat.pi / 2)
            
            var transformSignup = CGAffineTransform(scaleX: scaleSignup, y: scaleSignup)
            transformSignup = transformSignup.rotated(by: rotateAngleSignup)
            self.signupButton.transform = transformSignup
        }
        
    }
    
    
    // ===========================================================
    // MARK: - Keyboard
    // ===========================================================

    @objc func keyboarFrameChange(notification:NSNotification){
        
        let userInfo = notification.userInfo as! [String:AnyObject]
        
        // get top of keyboard in view
        let topOfKetboard = (userInfo[UIKeyboardFrameEndUserInfoKey] as! NSValue).cgRectValue .origin.y
        
        
        // get animation curve for animate view like keyboard animation
        var animationDuration:TimeInterval = 0.25
        var animationCurve:UIViewAnimationCurve = .easeOut
        if let animDuration = userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber {
            animationDuration = animDuration.doubleValue
        }
        
        if let animCurve = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber {
            animationCurve =  UIViewAnimationCurve.init(rawValue: animCurve.intValue)!
        }
        
        
        // check keyboard is showing
        let keyboardShow = topOfKetboard != self.view.frame.size.height
        
        
        //hide logo in little devices
        let hideLogo = self.view.frame.size.height < 667
        
        // set constraints
        backImageBottomConstraint.constant = self.view.frame.size.height - topOfKetboard
        
        logoTopConstraint.constant = keyboardShow ? (hideLogo ? 0:20):50
        logoHeightConstraint.constant = keyboardShow ? (hideLogo ? 0:40):60
        logoBottomConstraint.constant = keyboardShow ? 20:32
        logoButtomInSingupConstraint.constant = keyboardShow ? 20:32
        
        forgotPassTopConstraint.constant = keyboardShow ? 30:45
        
        signinButtonTopConstraint.constant = keyboardShow ? 25:30
        signupButtonTopConstraint.constant = keyboardShow ? 23:35
        
        signinButton.alpha = keyboardShow ? 1:0.7
        signupButton.alpha = keyboardShow ? 1:0.7
        
        
        
        // animate constraints changes
        UIView.beginAnimations(nil, context: nil)
        UIView.setAnimationDuration(animationDuration)
        UIView.setAnimationCurve(animationCurve)
        
        self.view.layoutIfNeeded()
        
        UIView.commitAnimations()
        
    }
    
    
    // ===========================================================
    // MARK: - Navigation
    // ===========================================================
    
    // In a storyboard-based application, you will often want to do a little preparation before navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if segue.destination is WebViewController {
            
            (segue.destination as! WebViewController).mode = (sender as! WebMode)
            
            if segue is CustomStoryboardSegue {
                (segue as! CustomStoryboardSegue).animationType = .swipeDown
            }
        } else if segue.destination is ResetPasswordViewController {
            
            (segue.destination as! ResetPasswordViewController).email = emailField.textFieldView.text
            
            if segue is CustomStoryboardSegue {
                (segue as! CustomStoryboardSegue).animationType = .swipeUp
            }
        }
    }
    
}


// ===========================================================
// MARK: - Google
// ===========================================================

extension SignInViewController: GIDSignInDelegate, GIDSignInUIDelegate {
    
    /**
     Verify user permissions with Google and login on the Eventhon system
     */
    @objc func loginGoogle() {
        GoogleAuth.signIn()
    }
    
    func sign(_ signIn: GIDSignIn!, present viewController: UIViewController!) {
        self.present(viewController, animated: true, completion: nil)
    }
    
    func sign(_ signIn: GIDSignIn!, dismiss viewController: UIViewController!) {
        self.dismiss(animated: true, completion: nil)
    }
    
    func sign(_ signIn: GIDSignIn!, didSignInFor user: GIDGoogleUser!, withError error: Error!) {
        if let error = error {
            if error._code == GIDSignInErrorCode.canceled.rawValue {
                self.showPopupAlert(title: NSLocalizedString("ERROR_HELP_TITLE", comment: ""),
                                    message: NSLocalizedString("ERROR_GOOGLE_SUBTITLE", comment: ""),
                                    image: nil,
                                    buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                    buttonColor: UIColor.primary.normal,
                                    style: .default,
                                    handler: nil)
            } else {
                self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                    message: error.localizedDescription,
                                    image: nil,
                                    buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                    buttonColor: UIColor.primary.normal,
                                    style: .default,
                                    handler: nil)
            }
        } else {
            let authentication = user.authentication
            let credential = GoogleAuthProvider.credential(withIDToken: (authentication?.idToken)!,
                                                           accessToken: (authentication?.accessToken)!)
            self.signin(with: credential)
        }
    }
    
}


// ===========================================================
// MARK: - Facebook
// ===========================================================

extension SignInViewController {
    
    /**
     Verify user permissions with Facebook and login on the Unuapp system
     */
    @objc func loginFacebook() {
        
        FacebookAuth.signIn(viewController: self) { (result,error) in
            if let error = error {
                self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                    message: error.localizedDescription,
                                    image: nil,
                                    buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                    buttonColor: UIColor.primary.normal,
                                    style: .default,
                                    handler: nil)
            } else if result!.isCancelled {
                self.showPopupAlert(title: NSLocalizedString("ERROR_HELP_TITLE", comment: ""),
                                    message: NSLocalizedString("ERROR_FACEBOOK_SUBTITLE", comment: ""),
                                    image: nil,
                                    buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                    buttonColor: UIColor.primary.normal,
                                    style: .default,
                                    handler: nil)
            } else {
                let credential = FacebookAuthProvider.credential(withAccessToken: FBSDKAccessToken.current().tokenString)
                
                self.signin(with: credential)
            }
        }
    }
}


// ===========================================================
// MARK: - Firebase
// ===========================================================

extension SignInViewController {
    
    /**
     Log in Firebase system with provider credential
     
     - parameter credential:    Provider credential.
     */
    func signin (with credential: AuthCredential) {
        
        UIApplication.shared.beginIgnoringInteractionEvents()
        activityIndicator.show(text: "")
        
        Auth.auth().signIn(with: credential) { (user, error) in
            UIApplication.shared.endIgnoringInteractionEvents()
            self.activityIndicator.hide()
            
            if let error = error {
                self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                    message: error.localizedDescription,
                                    image: nil,
                                    buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                    buttonColor: UIColor.primary.normal,
                                    style: .default,
                                    handler: nil)
            } else {
                self.performSegue(withIdentifier: "showMain", sender: nil)
            }
        }
    }
    
    /**
     Check all fields entered for the user and then try sign in on the system with email
     */
    func signin() {
        let email = emailField.textFieldView.text!
        let password = passwordField.textFieldView.text!
        
        // Array of errors detected in fields
        var errors = [String]()
        
        // Paint colors for error results
        let colorDefault = UIColor.clear
        let colorError = UIColor.red.withAlphaComponent(0.40)
        
        // Check email field
        if email.isEmpty {
            emailField.textFieldView.backgroundColor = colorError
            errors.append("The email can't be empty")
        } else if !email.isEmailValid {
            emailField.textFieldView.backgroundColor = colorError
            errors.append("The email have invalid sintax")
        } else {
            emailField.textFieldView.backgroundColor = colorDefault
        }
        
        // Check password field
        if password.isEmpty {
            passwordField.textFieldView.backgroundColor = colorError
            errors.append("The password can't be empty")
        } else {
            passwordField.textFieldView.backgroundColor = colorDefault
        }
        
        let numErrors = errors.count
        
        if numErrors > 0 {
            // Inform users about field errors
            showPopupAlert(title: NSLocalizedString("REVIEW_FORM", comment: ""),
                           message: errors.joined(separator: ".\n"),
                           image: nil,
                           buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                           buttonColor: UIColor.primary.normal,
                           style: .default,
                           handler: nil)
        } else {
            
            UIApplication.shared.beginIgnoringInteractionEvents()
            activityIndicator.show(text: "")
            
            Auth.auth().signIn(withEmail: email, password: password) { (user, error) in
                UIApplication.shared.endIgnoringInteractionEvents()
                self.activityIndicator.hide()
                
                if let error = error {
                    self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                        message: error.localizedDescription,
                                        image: nil,
                                        buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                        buttonColor: UIColor.primary.normal,
                                        style: .default,
                                        handler: nil)
                } else {
                    self.performSegue(withIdentifier: "showMain", sender: nil)
                }
            }
        }
    }
    
    /**
     Check all fields entered for the user and then try sign up on the system with email
     */
    func signup() {
        let email = emailFieldS.textFieldView.text!
        let password = passwordFieldS.textFieldView.text!
        
        // Array of errors detected in fields
        var errors = [String]()
        
        // Paint colors for error results
        let colorDefault = UIColor.clear
        let colorError = UIColor.red.withAlphaComponent(0.40)
        
        // Check email field
        if email.isEmpty {
            emailFieldS.textFieldView.backgroundColor = colorError
            errors.append("The email can't be empty")
        } else if !email.isEmailValid {
            emailFieldS.textFieldView.backgroundColor = colorError
            errors.append("The email has invalid sintax")
        } else {
            emailFieldS.textFieldView.backgroundColor = colorDefault
        }
        
        // Check password field
        if password.isEmpty {
            passwordFieldS.backgroundColor = colorError
            errors.append("The password can't be empty")
        } else if !password.isPasswordValid {
            passwordFieldS.backgroundColor = colorError
            errors.append("The password has to be 8 characters")
        } else {
            passwordFieldS.backgroundColor = colorDefault
        }
        
        let numErrors = errors.count
        
        if numErrors > 0 {
            // Inform users about field errors
            showPopupAlert(title: NSLocalizedString("REVIEW_FORM", comment: ""),
                           message: errors.joined(separator: ".\n"),
                           image: nil,
                           buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                           buttonColor: UIColor.primary.normal,
                           style: .default,
                           handler: nil)
        } else {
            
            UIApplication.shared.beginIgnoringInteractionEvents()
            activityIndicator.show(text: "")
            
            Auth.auth().createUser(withEmail: email, password: password) { (user, error) in
                UIApplication.shared.endIgnoringInteractionEvents()
                self.activityIndicator.hide()
            
                if let error = error  {
                    self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                        message: error.localizedDescription,
                                        image: nil,
                                        buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                        buttonColor: UIColor.primary.normal,
                                        style: .default,
                                        handler: nil)
                } else {
                    self.performSegue(withIdentifier: "showSignUpAvatar", sender: nil)
                }
            }
        }
    }
}



