/**
 SignupAvatarViewController.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import Firebase

class SignupAvatarViewController: BaseViewController {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    // Controller for select image source
    var imagePicker = UIImagePickerController()
    
    // Firebase storage reference
    let storageRef = Storage.storage().reference()
    
    // User photo
    var userPhoto: Photo?
    
    // Image tap gesture recognizer
    var tapRecognizerImage = UITapGestureRecognizer()
    
    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    // User photo image view
    @IBOutlet weak var imageViewPhoto: UIImageView!
    //  Email field
    @IBOutlet weak var nameField: AMInputView!
    
    
    // ===========================================================
    // MARK: - IB Actions
    // ===========================================================
    
    /**
     Start button action.
     */
    @IBAction func actionStart(_ sender: AnyObject) {
        startUp()
    }
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    override var prefersStatusBarHidden: Bool {
        return true
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        //  Add new functionality to view.
        self.view.hideKeyboardWhenTappedAround()
        
        // Init tap gesture recognizer to select new user image
        tapRecognizerImage.addTarget(self, action:#selector(self.showActionSheet))
        imageViewPhoto.isUserInteractionEnabled = true
        imageViewPhoto.addGestureRecognizer(tapRecognizerImage)
    }
    
    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        
        // Round image
        imageViewPhoto.layer.cornerRadius = 0.5 * imageViewPhoto.bounds.size.width
        imageViewPhoto.layer.masksToBounds = true
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        view.endEditing(true)
    }
    
    
    // ===========================================================
    // MARK: - OBJC Responds
    // ===========================================================
    
    /**
     Show action sheet for select image source
     */
    @objc func showActionSheet() {
        
        let optionMenu = UIAlertController(title: nil, message: "Change Profile Photo", preferredStyle: .actionSheet)
        
        let takePhotoAction = UIAlertAction(title: "Take a picture", style: .default, handler: fromCameraAction)
        let chooseLibraryAction = UIAlertAction(title: "Select existing photo", style: .default, handler: fromGalleryAction)
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: {
            (alert: UIAlertAction!) -> Void in
            print("Cancelled")
        })
        
        optionMenu.addAction(takePhotoAction)
        optionMenu.addAction(chooseLibraryAction)
        optionMenu.addAction(cancelAction)
        
        self.present(optionMenu, animated: true, completion: nil)
    }
    
    
    // ===========================================================
    // MARK: - Firebase
    // ===========================================================
    
    /**
     Start up new user properties and go to private views
     */
    func startUp() {
        
        UIApplication.shared.beginIgnoringInteractionEvents()
        activityIndicator.show(text: "")
        
        if userPhoto != nil {
            
            // Local file you want to upload
            let localFile = UIImageJPEGRepresentation(userPhoto!.thumbImage!, 0.8)
            // Create the file metadata
            let metadata = StorageMetadata()
            metadata.contentType = "image/jpg"
            // Create Filepath
            let filePath = "user/\(Auth.auth().currentUser!.uid)/profile"
            
            // Upload file and metadata
            storageRef.child(filePath).putData(localFile!, metadata: metadata){(metaData,error) in
                UIApplication.shared.endIgnoringInteractionEvents()
                self.activityIndicator.hide()
                
                if let error = error {
                    self.showPopupAlert(title: NSLocalizedString("ERROR_FRIENDLY", comment: ""),
                                        message: error.localizedDescription,
                                        image: nil,
                                        buttonLabel: NSLocalizedString("ACCEPT", comment: ""),
                                        buttonColor: UIColor.primary.normal,
                                        style: .default,
                                        handler: nil)
                }else{
                    //store downloadURL
                    self.userPhoto?.thumbUrl = metaData!.downloadURL()?.absoluteString
                    self.uploadUser()
                }
            }
            
        }
        else {
            UIApplication.shared.endIgnoringInteractionEvents()
            activityIndicator.hide()
            
            uploadUser()
        }
    }
    
    /**
     Upload firebase user
     */
    func uploadUser () {
        let name = nameField.textFieldView.text!
        
        let user = Auth.auth().currentUser
        
        if let user = user {
            let changeRequest = user.createProfileChangeRequest()
            
            if !name.isEmpty {
                changeRequest.displayName = name
            }
            if let thumbUrl = userPhoto?.thumbUrl {
                changeRequest.photoURL = URL(string: thumbUrl)
            }
            changeRequest.commitChanges { error in
                if let error = error {
                    // An error happened.
                    print(error.localizedDescription)
                } else {
                    // Profile updated.
                    self.performSegue(withIdentifier: "showMain", sender: self)
                }
            }
        }
    }
}


// ===========================================================
// MARK: - UIImagePickerController Delegate
// ===========================================================

extension SignupAvatarViewController: UIImagePickerControllerDelegate {
    
    /**
     Task to control the new image picked (from camera or gallery)
     */
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [String : Any]) {
        if let image = info[UIImagePickerControllerOriginalImage] as? UIImage {
            
            userPhoto = Photo(uiImage: image)
            
            imageViewPhoto.image = userPhoto?.image
        }
        
        dismiss(animated: true, completion: nil)
    }
    
    /**
     Task to select image from gallery
     */
    func fromGalleryAction(alert: UIAlertAction) {
        if UIImagePickerController.isSourceTypeAvailable(UIImagePickerControllerSourceType.photoLibrary) {
            imagePicker = UIImagePickerController()
            imagePicker.delegate = self
            imagePicker.sourceType = UIImagePickerControllerSourceType.photoLibrary;
            imagePicker.allowsEditing = true
            present(imagePicker, animated: true, completion: nil)
        }
    }
    
    /**
     Task to select image from camera
     */
    func fromCameraAction(alert: UIAlertAction) {
        if UIImagePickerController.isSourceTypeAvailable(UIImagePickerControllerSourceType.camera) {
            imagePicker = UIImagePickerController()
            imagePicker.delegate = self
            imagePicker.sourceType = UIImagePickerControllerSourceType.camera;
            imagePicker.allowsEditing = false
            present(imagePicker, animated: true, completion: nil)
        }
    }
}


// ===========================================================
// MARK: - UINavigationController Delegate
// ===========================================================

extension SignupAvatarViewController: UINavigationControllerDelegate {
    
}

