/**
 AMInputView.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit

protocol AMInputDelegate {
    
    func amInputView(hasText: Bool)
    
}

@IBDesignable class AMInputView: UIView {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    var delegate: AMInputDelegate?
    
    let animationDuration = 0.25

    
    // ===========================================================
    // MARK: - IB Outlets
    // ===========================================================
    
    @IBOutlet weak var contentView: UIView!
    @IBOutlet weak var textFieldView: UITextField!
    @IBOutlet weak var backWhiteView: UIView!
    @IBOutlet weak var labelView: UILabel!
    
    // Constraints
    @IBOutlet weak var backViewHeightExpandConstraint: NSLayoutConstraint!
    @IBOutlet weak var labelViewRightExpandConstraint: NSLayoutConstraint!
    @IBOutlet weak var labelViewLeftExpandConstraint: NSLayoutConstraint!
    @IBOutlet weak var labelViewHTopExpandConstraint: NSLayoutConstraint!
    
    // ===========================================================
    // MARK: - IB Inspectables
    // ===========================================================
    
    @IBInspectable var Title:String = ""{
        didSet{
            if labelView != nil{
                labelView.text = Title
            }
        }
    }
    
    @IBInspectable var KeyboardType:Int = UIKeyboardType.default.rawValue{
        didSet{
            if textFieldView != nil{
                textFieldView.keyboardType = UIKeyboardType(rawValue: KeyboardType)!
            }
        }
    }
    
    @IBInspectable var isSecureTextEntry:Bool = false{
        didSet{
            if textFieldView != nil{
                textFieldView.isSecureTextEntry = isSecureTextEntry
            }
        }
    }
    
    @IBInspectable var autocorrectionType:Int = UITextAutocorrectionType.default.rawValue{
        didSet{
            if textFieldView != nil{
                textFieldView.autocorrectionType = UITextAutocorrectionType(rawValue: autocorrectionType)!
            }
        }
    }
    
    @IBInspectable var returnKeyType:Int = UIReturnKeyType.default.rawValue{
        didSet{
            if textFieldView != nil{
                textFieldView.returnKeyType = UIReturnKeyType(rawValue: returnKeyType)!
            }
        }
    }
    
    
    // ===========================================================
    // MARK: - View Lifecycle
    // ===========================================================
    
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        setup()
    }
    
    override init(frame: CGRect) {
        super.init(frame: frame)
        setup()
    }
    
    private func setup(){
        
        loadNib()
        
        self.addSubview(contentView)
        
        self.addConstraints([
            
            NSLayoutConstraint(item: contentView, attribute: .width, relatedBy: .equal, toItem: self, attribute: .width, multiplier: 1, constant: 0),
            NSLayoutConstraint(item: contentView, attribute: .height, relatedBy: .equal, toItem: self, attribute: .height, multiplier: 1, constant: 0),
            NSLayoutConstraint(item: contentView, attribute: .centerX, relatedBy: .equal, toItem: self, attribute: .centerX, multiplier: 1, constant: 0),
            NSLayoutConstraint(item: contentView, attribute: .centerY, relatedBy: .equal, toItem: self, attribute: .centerY, multiplier: 1, constant: 0)
            
            ])
        
        
        
        textFieldView.tintColor = UIColor.white
        textFieldView.delegate = self
        textFieldView.keyboardAppearance = .dark
    }
    
    private func loadNib() {
        let bundle = Bundle(for: AMInputView.classForCoder())
        bundle.loadNibNamed("AMInputView", owner: self, options: nil)
    }
    
    
    // ===========================================================
    // MARK: - Change view
    // ===========================================================

    func setViewExpandMode(expand:Bool){
        
        backViewHeightExpandConstraint.priority = expand ? UILayoutPriority(500):UILayoutPriority(200)
        labelViewHTopExpandConstraint.priority = expand ? UILayoutPriority(500):UILayoutPriority(200)
        labelViewRightExpandConstraint.priority = expand ? UILayoutPriority(500):UILayoutPriority(300)
        labelViewLeftExpandConstraint.priority = expand ? UILayoutPriority(999):UILayoutPriority(300)
        
        UIView.animate(withDuration: animationDuration) {
            self.layoutIfNeeded()
            self.backWhiteView.alpha = expand ? 0.6:0.08
            
            let scale:CGFloat = expand ? 0.7:1
            
            self.labelView.transform = CGAffineTransform(scaleX: scale, y: scale)
            
            if expand {
                
                let leftMargin = (-1 * ((1-scale)*(self.labelView.layer.bounds.size.width))/2 ) - 11;
                self.labelView.transform = self.labelView.transform.translatedBy(x: leftMargin, y: 0)
            }
        }
    }
    
}

// ===========================================================
// MARK: - UITextField Delegate
// ===========================================================

extension AMInputView: UITextFieldDelegate {
    
    func textFieldDidBeginEditing(_ textField: UITextField) {
        
        if (textField.text?.lengthOfBytes(using: .utf8))! < 1 {
            setViewExpandMode(expand:true)
        }
    }
    
    func textFieldDidEndEditing(_ textField: UITextField) {
        
        if (textField.text?.lengthOfBytes(using: .utf8))! < 1 {
            setViewExpandMode(expand:false)
        }
    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange, replacementString string: String) -> Bool {
        
        let textLength = textField.text!.count + string.count - range.length
        
        delegate?.amInputView(hasText: textLength > 0)
        
        return true
    }
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        
        self.endEditing(true)
        return true
    }
}


