/**
 CustomStoryboardSegue.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit


enum CustomSegueAnimation {
  case swipeRight
  case swipeLeft
  case swipeDown
  case swipeUp
  case growScale
}


/**
 Custom storyboard segues to use in transictions between to views.
 */
class CustomStoryboardSegue: UIStoryboardSegue {
  
  // Default
  var animationType = CustomSegueAnimation.swipeRight
  
  /**
   Select animation type.
   */
  override func perform() {
    switch animationType {
    case .swipeRight:
      animateSwipeRight()
    case .swipeLeft:
      animateSwipeLeft()
    case .swipeDown:
      animateSwipeDown()
    case .swipeUp:
      animateSwipeUp()
    case .growScale:
      animateGrowScale()
    }
  }
  
  /**
   Animate segue swipe right.
   */
  private func animateSwipeRight() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:screenWidth, y:0.0, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Add new view controller into navigation stack
    var vcs = fromVC.navigationController?.viewControllers
    vcs?.append(toVC)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: -screenWidth, dy: 0.0)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: -screenWidth, dy: 0.0)
      
    }) { (Finished) -> Void in
      fromVC.navigationController?.setViewControllers(vcs!,
                                                      animated: false)
    }
  }
  
  /**
   Animate segue swipe left.
   */
  private func animateSwipeLeft() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:-screenWidth, y:0.0, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Add new view controller into navigation stack
    var vcs = fromVC.navigationController?.viewControllers
    vcs?.append(toVC)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: screenWidth, dy: 0.0)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: screenWidth, dy: 0.0)
      
    }) { (Finished) -> Void in
      fromVC.navigationController?.setViewControllers(vcs!,
                                                      animated: false)
    }
  }
  
  /**
   Animate segue swipe down.
   */
  private func animateSwipeDown() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:0.0, y:screenHeight, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Add new view controller into navigation stack
    var vcs = fromVC.navigationController?.viewControllers
    vcs?.append(toVC)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: 0.0, dy: -screenHeight)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: 0.0, dy: -screenHeight)
      
    }) { (Finished) -> Void in
      fromVC.navigationController?.setViewControllers(vcs!,
                                                      animated: false)
    }
  }
  
  /**
   Animate segue swipe up.
   */
  private func animateSwipeUp() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:0.0, y:-screenHeight, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Add new view controller into navigation stack
    var vcs = fromVC.navigationController?.viewControllers
    vcs?.append(toVC)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: 0.0, dy: screenHeight)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: 0.0, dy: screenHeight)
      
    }) { (Finished) -> Void in
      fromVC.navigationController?.setViewControllers(vcs!,
                                                      animated: false)
    }
  }
  
  /**
   Animate segue grow scale.
   */
  private func animateGrowScale() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Specify the initial position of the destination view.
    secondVCView?.transform = CGAffineTransform(scaleX: 0.05, y: 0.05)
    secondVCView?.center = firstVCView!.center
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Add new view controller into navigation stack
    var vcs = fromVC.navigationController?.viewControllers
    vcs?.append(toVC)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      secondVCView?.transform = CGAffineTransform(scaleX: 1.0, y: 1.0)
      
    }) { (Finished) -> Void in
      fromVC.navigationController?.setViewControllers(vcs!,
                                                      animated: false)
    }
  }
}


/**
 Custom storyboard unwind segues to use in transictions between to views.
 */
class CustomUnwindSegue: UIStoryboardSegue {
  
  // Default
  var animationType: CustomSegueAnimation = .swipeRight
  
  /**
   Select animation type.
   */
  override func perform() {
    switch animationType {
    case .swipeRight:
      animateSwipeRight()
    case .swipeLeft:
      animateSwipeLeft()
    case .swipeDown:
      animateSwipeDown()
    case .swipeUp:
      animateSwipeUp()
    case .growScale:
      animateGrowScale()
    }
  }
  
  /**
   Animate unwind segue swipe right.
   */
  private func animateSwipeRight() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:-screenWidth, y:0.0, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: screenWidth, dy: 0.0)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: screenWidth, dy: 0.0)
      
    }) { (Finished) -> Void in
      fromVC.navigationController!.popViewController(animated: false)
    }
  }
  
  /**
   Animate unwind segue swipe left.
   */
  private func animateSwipeLeft() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:screenWidth, y:0.0, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: -screenWidth, dy: 0.0)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: -screenWidth, dy: 0.0)
      
    }) { (Finished) -> Void in
      fromVC.navigationController!.popViewController(animated: false)
    }
  }
  
  /**
   Animate unwind segue swipe down.
   */
  private func animateSwipeDown() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:0.0, y:-screenHeight, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: 0.0, dy: screenHeight)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: 0.0, dy: screenHeight)
      
    }) { (Finished) -> Void in
      fromVC.navigationController!.popViewController(animated: false)
    }
  }
  
  /**
   Animate unwind segue swipe up.
   */
  private func animateSwipeUp() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Get the screen width and height.
    let screenWidth = UIScreen.main.bounds.size.width
    let screenHeight = UIScreen.main.bounds.size.height
    
    // Specify the initial position of the destination view.
    secondVCView?.frame = CGRect(x:0.0, y:screenHeight, width:screenWidth, height:screenHeight)
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.frame = firstVCView!.frame.offsetBy(dx: 0.0, dy: -screenHeight)
      secondVCView?.frame = secondVCView!.frame.offsetBy(dx: 0.0, dy: -screenHeight)
      
    }) { (Finished) -> Void in
      fromVC.navigationController!.popViewController(animated: false)
    }
  }
  
  /**
   Animate unwind segue grow scale.
   */
  private func animateGrowScale() {
    
    // Assign the source and destination view controllers to local variables.
    let fromVC = self.source
    let toVC = self.destination
    
    // Assign the source and destination views to local variables.
    let firstVCView = fromVC.view as UIView!
    let secondVCView = toVC.view as UIView!
    
    // Specify the initial position of the destination view.
    secondVCView?.transform = CGAffineTransform(scaleX: 1.0, y: 1.0)
    secondVCView?.center = firstVCView!.center
    
    // Access the app's key window and insert the destination view above the current (source) one.
    let window = UIApplication.shared.keyWindow
    window?.insertSubview(secondVCView!, aboveSubview: firstVCView!)
    
    // Animate the transition.
    UIView.animate(withDuration: 0.4, animations: { () -> Void in
      firstVCView?.transform = CGAffineTransform(scaleX: 0.05, y: 0.05)
      
    }) { (Finished) -> Void in
      fromVC.navigationController!.popViewController(animated: false)
    }
  }
}


