/**
 Extensions.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import UIKit
import SystemConfiguration
import MapKit
import PopupDialog
import SwiftMessages


// ===========================================================
// MARK: - Bundle
// ===========================================================

extension Bundle {
    var releaseVersionNumber: String? {
        return infoDictionary?["CFBundleShortVersionString"] as? String
    }
    var buildVersionNumber: String? {
        return infoDictionary?["CFBundleVersion"] as? String
    }
}


// ===========================================================
// MARK: - CAShapeLayer
// ===========================================================

extension CAShapeLayer {
    
    /**
     Draw circle.
     
     - parameter location: Location on the screen.
     - parameter withRadius: Radius of the circle.
     - parameter andColor: Color of the circle.
     - parameter filled: If is filled or not.
     */
    func drawCircleAtLocation(location: CGPoint, withRadius radius: CGFloat, andColor color: UIColor, filled: Bool) {
        fillColor = filled ? color.cgColor : UIColor.white.cgColor
        strokeColor = color.cgColor
        let origin = CGPoint(x: location.x - radius, y: location.y - radius)
        path = UIBezierPath(ovalIn: CGRect(origin: origin, size: CGSize(width: radius * 2, height: radius * 2))).cgPath
    }
}


// ===========================================================
// MARK: - CLLocationDistance
// ===========================================================

extension CLLocationDistance {
    
    func toKilometres() -> CLLocationDistance {
        
        var targetDistance = self
        targetDistance =  targetDistance*0.00062137
        return targetDistance
    }
    
}


// ===========================================================
// MARK: - Date
// ===========================================================

extension Date {
    
    init(mysqlTimestamp: String){
        let dateFormatter = DateFormatter()
        dateFormatter.dateFormat = "YYYY-MM-dd HH:mm:ss"
        self = dateFormatter.date(from: mysqlTimestamp)!
    }
    
    /**
     Add style to date.
     
     - parameter style: Style.
     
     - returns: String with style.
     */
    func dateStringWith(style: DateFormatter.Style) -> String {
        let dateFormatter = DateFormatter()
        dateFormatter.dateStyle = style
        return dateFormatter.string(from: self)
    }
    
    /**
     Add custom style to date.
     
     - parameter style: Custom style.
     
     - returns: String with custom style.
     */
    func dateStringWith(dateFormat: String) -> String {
        let dateFormatter = DateFormatter()
        dateFormatter.dateFormat = dateFormat
        return dateFormatter.string(from: self)
    }
    
    /// Returns the amount of years from another date
    func years(from date: Date) -> Int {
        return Calendar.current.dateComponents([.year], from: date, to: self).year ?? 0
    }
    /// Returns the amount of months from another date
    func months(from date: Date) -> Int {
        return Calendar.current.dateComponents([.month], from: date, to: self).month ?? 0
    }
    /// Returns the amount of weeks from another date
    func weeks(from date: Date) -> Int {
        return Calendar.current.dateComponents([.weekOfYear], from: date, to: self).weekOfYear ?? 0
    }
    /// Returns the amount of days from another date
    func days(from date: Date) -> Int {
        return Calendar.current.dateComponents([.day], from: date, to: self).day ?? 0
    }
    /// Returns the amount of hours from another date
    func hours(from date: Date) -> Int {
        return Calendar.current.dateComponents([.hour], from: date, to: self).hour ?? 0
    }
    /// Returns the amount of minutes from another date
    func minutes(from date: Date) -> Int {
        return Calendar.current.dateComponents([.minute], from: date, to: self).minute ?? 0
    }
    /// Returns the amount of seconds from another date
    func seconds(from date: Date) -> Int {
        return Calendar.current.dateComponents([.second], from: date, to: self).second ?? 0
    }
    /// Returns the a custom time interval description from another date
    func offset(from date: Date) -> String {
        if years(from: date)   > 0 { return "\(years(from: date))y"   }
        if months(from: date)  > 0 { return "\(months(from: date))M"  }
        if weeks(from: date)   > 0 { return "\(weeks(from: date))w"   }
        if days(from: date)    > 0 { return "\(days(from: date))d"    }
        if hours(from: date)   > 0 { return "\(hours(from: date))h"   }
        if minutes(from: date) > 0 { return "\(minutes(from: date))m" }
        if seconds(from: date) > 0 { return "\(seconds(from: date))s" }
        return ""
    }
}


// ===========================================================
// MARK: - DateFormatter
// ===========================================================

extension DateFormatter {
    
    /**
     New init.
     
     - Parameter dateStyle: Create new DateFormatter with defined Style.
     */
    convenience init(dateStyle: DateFormatter.Style) {
        self.init()
        self.dateStyle = dateStyle
    }
}


// ===========================================================
// MARK: - Double
// ===========================================================

extension Double {
    
    /**
     Given a value to round and a factor to round to,
     round the value UP to the largest previous multiple
     of that factor.
     
     - parameter toNearest: Factor to round.
     - return: Rounded value.
     */
    func roundUp (toNearest: Double) -> Double {
    
        return ceil(self / toNearest) * toNearest
    }
    
    /**
     Given a value to round and a factor to round to,
     round the value DOWN to the largest previous multiple
     of that factor.
     
     - parameter toNearest: Factor to round.
     - return: Rounded value.
     */
    func roundDown(toNearest: Double) -> Double {
        return floor(self / toNearest) * toNearest
    }
    
    /**
     Given a value to round,
     round the value DOWN to the largest previous multiple
     of an exponential factor.
     
     - return: Rounded value.
     */
    func roundDownExp() -> Double {
        switch self {
        case 0...10:
            return self.roundDown(toNearest: 1)
        case 10...100:
            return self.roundDown(toNearest: 5)
        default:
            return self.roundDown(toNearest: 50)
        }
    }
}


// ===========================================================
// MARK: - Float
// ===========================================================

extension Float {
    
    /**
     Transform a float to a string value format (5, 5.1, 5.2 ...)
     */
    var cleanValue: String {
        return String(format: self == floor(self) ? "%.0f" : "%.1f", self)
    }
}


// ===========================================================
// MARK: - Locale
// ===========================================================

extension Locale {
    
    /**
     Get current country name.
     
     - return: Current country code.
     */
    static func getCurrentCountry() -> String? {
        let id = self.identifier(fromComponents: [NSLocale.Key.countryCode.rawValue: self.current.regionCode!])
        let country = NSLocale(localeIdentifier: self.current.languageCode!).displayName(forKey: NSLocale.Key.identifier, value: id) ?? ""
        
        return country
    }
    
    /**
     Get country code from country name.
     
     - parameter countryName: Country name.
     
     - return: Country code.
     */
    static func countryCodeFrom(countryName:String) -> String?
    {
        return self.isoRegionCodes.filter{self.countryNameFrom(countryCode: $0) == countryName}.first
    }
    
    /**
     Get country name from country code.
     
     - parameter countryCode: Country code.
     
     - return: Country name.
     */
    static func countryNameFrom(countryCode : String) -> String
    {
        return NSLocale(localeIdentifier: self.current.languageCode!).displayName(forKey: NSLocale.Key.countryCode, value: countryCode) ?? ""
    }
    
    /**
     Get country flag unicode from country code.
     
     - parameter countryCode: Country code.
     
     - return: Country flag unicode.
     */
    static func flagFrom(countryCode:String) -> String {
        let base : UInt32 = 127397
        var s = ""
        for v in countryCode.unicodeScalars {
            s.unicodeScalars.append(UnicodeScalar(base + v.value)!)
        }
        return String(s)
    }
}


// ===========================================================
// MARK: - MKMapView
// ===========================================================

extension MKMapView {
    
    /**
     Zoom map to user location.
     
     - parameter distanceLat: The amount of north-to-south distance mesured in meters.
     - parameter distanceLon: The amount of east-to-west distance mesured in meters.
     */
    func zoomToUserLocationWith(distanceLat:CLLocationDistance, distanceLon:CLLocationDistance) {
        guard let coordinate = userLocation.location?.coordinate else { return }
        let region = MKCoordinateRegionMakeWithDistance(coordinate, distanceLat, distanceLon)
        setRegion(region, animated: true)
    }
}


// ===========================================================
// MARK: - Sequence
// ===========================================================

public extension Sequence {
    func categorise<U : Hashable>(_ key: (Iterator.Element) -> U) -> [U:[Iterator.Element]] {
        var dict: [U:[Iterator.Element]] = [:]
        for el in self {
            let key = key(el)
            if case nil = dict[key]?.append(el) { dict[key] = [el] }
        }
        return dict
    }
}


// ===========================================================
// MARK: - NSObject
// ===========================================================


protocol Utilities {
}

extension NSObject: Utilities{
    
    enum ReachabilityStatus {
        case notReachable
        case reachableViaWWAN
        case reachableViaWiFi
    }
    
    var currentReachabilityStatus: ReachabilityStatus {
        
        var zeroAddress = sockaddr_in()
        zeroAddress.sin_len = UInt8(MemoryLayout<sockaddr_in>.size)
        zeroAddress.sin_family = sa_family_t(AF_INET)
        
        guard let defaultRouteReachability = withUnsafePointer(to: &zeroAddress, {
            $0.withMemoryRebound(to: sockaddr.self, capacity: 1) {
                SCNetworkReachabilityCreateWithAddress(nil, $0)
            }
        }) else {
            return .notReachable
        }
        
        var flags: SCNetworkReachabilityFlags = []
        if !SCNetworkReachabilityGetFlags(defaultRouteReachability, &flags) {
            return .notReachable
        }
        
        if flags.contains(.reachable) == false {
            // The target host is not reachable.
            return .notReachable
        }
        else if flags.contains(.isWWAN) == true {
            // WWAN connections are OK if the calling application is using the CFNetwork APIs.
            return .reachableViaWWAN
        }
        else if flags.contains(.connectionRequired) == false {
            // If the target host is reachable and no connection is required then we'll assume that you're on Wi-Fi...
            return .reachableViaWiFi
        }
        else if (flags.contains(.connectionOnDemand) == true || flags.contains(.connectionOnTraffic) == true) && flags.contains(.interventionRequired) == false {
            // The connection is on-demand (or on-traffic) if the calling application is using the CFSocketStream or higher APIs and no [user] intervention is needed
            return .reachableViaWiFi
        }
        else {
            return .notReachable
        }
    }
    
}


// ===========================================================
// MARK: - String
// ===========================================================

extension String {
    
    /**
     Check if email is valid.
     */
    var isEmailValid: Bool {
        return Regex.checkRegexInText(regex: "[A-Z0-9a-z._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,64}",
                                      text: self)
    }
    
    /**
     Check if password is valid.
     */
    var isPasswordValid: Bool {
        return Regex.checkPasswordComplexity(password: self, length: 8, patternsToEscape: ["1234","qwer","asdf"], caseSensitivty: true, numericDigits: true)
    }
    
    /**
     Replace substring for other replacement string.
     
     - parameter string:        Old substring.
     - parameter replacement:   Replacement substring.
     
     - returns: String modified.
     */
    func replace(_ string:String, replacement:String) -> String {
        return self.replacingOccurrences(of: string, with: replacement, options: NSString.CompareOptions.literal, range: nil)
    }
    
    /**
     Remove white spaces.
     
     - returns: String modified.
     */
    func removeWhitespace() -> String {
        return self.replace(" ", replacement: "")
    }
}


// ===========================================================
// MARK: - UIApplication
// ===========================================================

private var firstLaunch : Bool = false
extension UIApplication {
    
    /**
     Check for is first launch.
     
     - return: Only true on first invocation after app install, false on all further invocations
     */
    static func isFirstLaunch() -> Bool {
        let firstLaunchFlag = "isFirstLaunchFlag"
        let isFirstLaunch = UserDefaults.standard.string(forKey: firstLaunchFlag) == nil
        if (isFirstLaunch) {
            firstLaunch = isFirstLaunch
            UserDefaults.standard.set("false", forKey: firstLaunchFlag)
            UserDefaults.standard.synchronize()
        }
        return firstLaunch || isFirstLaunch
    }
}


// ===========================================================
// MARK: - UIBarButtonItem
// ===========================================================

private var handle: UInt8 = 0
extension UIBarButtonItem {
    
    /**
     Badge layer.
     */
    private var badgeLayer: CAShapeLayer? {
        if let b: AnyObject = objc_getAssociatedObject(self, &handle) as AnyObject? {
            return b as? CAShapeLayer
        } else {
            return nil
        }
    }
    
    /**
     Add badge.
     
     - parameter number: Number inside the badge.
     - parameter withOffset: Offset of the badge.
     - parameter andColor: Color of the badge.
     - parameter andFilled: If is filled or not.
     */
    func addBadge(number: Int, withOffset offset: CGPoint = CGPoint.zero, andColor color: UIColor = UIColor.red, andFilled filled: Bool = true) {
        guard let view = self.value(forKey: "view") as? UIView else { return }
        
        badgeLayer?.removeFromSuperlayer()
        
        var badgeWidth = 8
        var numberOffset = 4
        
        if number > 9 {
            badgeWidth = 12
            numberOffset = 6
        }
        
        // Initialize Badge
        let badge = CAShapeLayer()
        let radius = CGFloat(7)
        let location = CGPoint(x: view.frame.width - (radius + offset.x), y: (radius + offset.y))
        badge.drawCircleAtLocation(location: location, withRadius: radius, andColor: color, filled: filled)
        view.layer.addSublayer(badge)
        
        // Initialiaze Badge's label
        let label = CATextLayer()
        label.string = "\(number)"
        label.alignmentMode = kCAAlignmentCenter
        label.fontSize = 11
        label.frame = CGRect(origin: CGPoint(x: location.x - CGFloat(numberOffset), y: offset.y), size: CGSize(width: badgeWidth, height: 16))
        label.foregroundColor = filled ? UIColor.white.cgColor : color.cgColor
        label.backgroundColor = UIColor.clear.cgColor
        label.contentsScale = UIScreen.main.scale
        badge.addSublayer(label)
        
        // Save Badge as UIBarButtonItem property
        objc_setAssociatedObject(self, &handle, badge, .OBJC_ASSOCIATION_RETAIN_NONATOMIC)
    }
    
    /**
     Update number of the badge.
     
     - parameter number: Number inside the badge.
     */
    func updateBadge(number: Int) {
        if let text = badgeLayer?.sublayers?.filter({ $0 is CATextLayer }).first as? CATextLayer {
            text.string = "\(number)"
        }
    }
    
    /**
     Remove badge.
     */
    func removeBadge() {
        badgeLayer?.removeFromSuperlayer()
    }
}


// ===========================================================
// MARK: - UIButton
// ===========================================================

extension UIButton {
    
    /**
     Add a border width functionality.
     */
    @IBInspectable var borderWidth: CGFloat {
        get {
            return layer.borderWidth
        }
        set {
            layer.borderWidth = newValue
            layer.masksToBounds = newValue > 0
        }
    }
    
    /**
     Add a border color functionality.
     */
    @IBInspectable var borderColor: UIColor {
        get {
            return UIColor(cgColor: layer.borderColor!)
        }
        set {
            layer.borderColor = newValue.cgColor
        }
    }
    
    /**
     Add a shadow button.
     */
    func setShadowButton(){
        layer.shadowColor = UIColor.black.cgColor
        layer.shadowOffset = CGSize(width: 0.0, height: 2.0)
        layer.masksToBounds = false
        layer.shadowRadius = 1.0
        layer.shadowOpacity = 0.5
    }
}


// ===========================================================
// MARK: - UIColor
// ===========================================================

extension UIColor {
    
    // Structs
    
    struct Tone {
        var light: UIColor?
        var normal: UIColor?
        var dark: UIColor?
        var accent: UIColor?
    }
    
    // Color Style Variables
    
    class var primary: Tone {
        var tone = Tone()
        
        tone.light = UIColor(withHexValue: 0xB2DFDB)
        tone.normal = UIColor(withHexValue: 0x009688)
        tone.dark = UIColor(withHexValue: 0x00695C)
        tone.accent = UIColor(withHexValue: 0x1DE9B6)
        
        return tone
    }
    
    class var secondary: Tone {
        var tone = Tone()
        
        tone.light = UIColor(withHexValue: 0xFFFFFF)
        tone.normal = UIColor(withHexValue: 0xFFFFFF)
        tone.dark = UIColor(withHexValue: 0xFFFFFF)
        tone.accent = UIColor(withHexValue: 0xFFFFFF)
        
        return tone
    }
    
    // Constructors
    
    /**
     Init with hexadecimal and alpha values.
     
     - parameter withHexValue: Hexadecimal integer.
     - parameter alpha: Alpha component.
     */
    convenience init(withHexValue hex: Int, andAlpha alpha:CGFloat = 1.0){
        self.init(
            red: CGFloat((hex & 0xFF0000) >> 16) / 255.0,
            green: CGFloat((hex & 0x00FF00) >> 8) / 255.0,
            blue: CGFloat(hex & 0x0000FF) / 255.0,
            alpha: alpha
        )
    }
    
    /**
     Init with string agbr hexadecimal encoded.
     
     - parameter hex: Hexadeximal agbr encoded string.
     */
    convenience init(string: String) {
        let scanner = Scanner(string: string)
        var color: UInt32 = 0
        scanner.scanHexInt32(&color)
        
        let alpha = CGFloat((color >> 24) & 0x000000FF) / 255.0
        let blue = CGFloat((color >> 16) & 0x000000FF) / 255.0
        let green = CGFloat((color >> 8) & 0x000000FF) / 255.0
        let red = CGFloat(color & 0x000000FF) / 255.0
        
        self.init(
            red: red,
            green: green,
            blue: blue,
            alpha: alpha)
    }
}


// ===========================================================
// MARK: - UIDevice
// ===========================================================

extension UIDevice {
    
    /**
     Get model name to control users priorities.
     */
    var modelName: String {
        var systemInfo = utsname()
        uname(&systemInfo)
        let machineMirror = Mirror(reflecting: systemInfo.machine)
        let identifier = machineMirror.children.reduce("") { identifier, element in
            guard let value = element.value as? Int8 , value != 0 else { return identifier }
            return identifier + String(UnicodeScalar(UInt8(value)))
        }
        
        switch identifier {
        case "iPod5,1":                                 return "iPod Touch 5"
        case "iPod7,1":                                 return "iPod Touch 6"
        case "iPhone3,1", "iPhone3,2", "iPhone3,3":     return "iPhone 4"
        case "iPhone4,1":                               return "iPhone 4s"
        case "iPhone5,1", "iPhone5,2":                  return "iPhone 5"
        case "iPhone5,3", "iPhone5,4":                  return "iPhone 5c"
        case "iPhone6,1", "iPhone6,2":                  return "iPhone 5s"
        case "iPhone7,2":                               return "iPhone 6"
        case "iPhone7,1":                               return "iPhone 6 Plus"
        case "iPhone8,1":                               return "iPhone 6s"
        case "iPhone8,2":                               return "iPhone 6s Plus"
        case "iPhone9,1", "iPhone9,3":                  return "iPhone 7"
        case "iPhone9,2", "iPhone9,4":                  return "iPhone 7 Plus"
        case "iPhone8,4":                               return "iPhone SE"
        case "iPad2,1", "iPad2,2", "iPad2,3", "iPad2,4":return "iPad 2"
        case "iPad3,1", "iPad3,2", "iPad3,3":           return "iPad 3"
        case "iPad3,4", "iPad3,5", "iPad3,6":           return "iPad 4"
        case "iPad4,1", "iPad4,2", "iPad4,3":           return "iPad Air"
        case "iPad5,3", "iPad5,4":                      return "iPad Air 2"
        case "iPad2,5", "iPad2,6", "iPad2,7":           return "iPad Mini"
        case "iPad4,4", "iPad4,5", "iPad4,6":           return "iPad Mini 2"
        case "iPad4,7", "iPad4,8", "iPad4,9":           return "iPad Mini 3"
        case "iPad5,1", "iPad5,2":                      return "iPad Mini 4"
        case "iPad6,3", "iPad6,4", "iPad6,7", "iPad6,8":return "iPad Pro"
        case "AppleTV5,3":                              return "Apple TV"
        case "i386", "x86_64":                          return "Simulator"
        default:                                        return identifier
        }
    }
}


// ===========================================================
// MARK: - UIImage
// ===========================================================

extension UIImage {
    
    func resizeImage(size:CGSize) -> UIImage? {
        
        UIGraphicsBeginImageContext(size)
        self.draw(in: CGRect(x:0, y:0, width:size.width, height:size.height))
        let resizedImage = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        
        return resizedImage
    }
    
}


// ===========================================================
// MARK: - UIProgressView
// ===========================================================

extension UIProgressView {
    
    /**
     Transform bar height from progress view height
     */
    @IBInspectable var barHeight : CGFloat {
        get {
            return transform.d * 2.0
        }
        set {
            let heightScale = newValue / 2.0
            let c = center
            transform = CGAffineTransform(scaleX: 1.0, y: heightScale)
            center = c
        }
    }
}


// ===========================================================
// MARK: - UISearchBar
// ===========================================================

extension UISearchBar{
    
    /**
     Add a text color functionality.
     */
    @IBInspectable var textColor: UIColor {
        get {
            return self.textColor
        }
        set {
            let searchTextField: UITextField? = self.value(forKey: "searchField") as? UITextField
            searchTextField?.textColor = newValue
        }
    }
    
    /**
     Add a place holder text color functionality.
     */
    @IBInspectable var placeHolderColor: UIColor {
        get {
            return self.placeHolderColor
        }
        set {
            let searchTextField: UITextField? = self.value(forKey: "searchField") as? UITextField
            searchTextField?.placeHolderColor = newValue
        }
    }
    
    /**
     Add a search image functionality.
     */
    @IBInspectable var searchImage: UIImage {
        get {
            return self.searchImage
        }
        set {
            self.setImage(newValue, for: .search, state: .normal)
        }
    }
    
    /**
     Add a clear image functionality.
     */
    @IBInspectable var clearImage: UIImage {
        get {
            return self.clearImage
        }
        set {
            self.setImage(newValue, for: .clear, state: .normal)
        }
    }
}


// ===========================================================
// MARK: - UITextField
// ===========================================================

extension UITextField{
    
    /**
     Add a border width functionality.
     */
    @IBInspectable var borderWidth: CGFloat {
        get {
            return layer.borderWidth
        }
        set {
            layer.borderWidth = newValue
            layer.masksToBounds = newValue > 0
        }
    }
    
    /**
     Add a border color functionality.
     */
    @IBInspectable var borderColor: UIColor {
        get {
            return UIColor(cgColor: layer.borderColor!)
        }
        set {
            layer.borderColor = newValue.cgColor
        }
    }
    
    /**
     Add a place holder text color functionality.
     */
    @IBInspectable var placeHolderColor: UIColor {
        get {
            return self.placeHolderColor
        }
        set {
            self.attributedPlaceholder = NSAttributedString(string:self.placeholder != nil ? self.placeholder! : "", attributes:[NSAttributedStringKey.foregroundColor: newValue])
        }
    }
    
    /**
     Check if a field has text while it is editing.
     
     - parameter textField:     UITextField to be checked.
     - parameter newCharacter:  It is the last entered character when UITextField is being editing.
     
     - returns: UITextField has text.
     */
    func hasTextWhileEditing (_ textField: UITextField, newCharacter: String) -> Bool {
        
        var hasText = false
        
        if newCharacter != "" {
            hasText = true
        } else if textField.text!.count > 1 {
            hasText = true
        } else {
            hasText = false
        }
        return hasText
    }
}


// ===========================================================
// MARK: - UITextView
// ===========================================================

extension UITextView {
    
    /**
     Add a border width functionality.
     */
    @IBInspectable var borderWidth: CGFloat {
        get {
            return layer.borderWidth
        }
        set {
            layer.borderWidth = newValue
            layer.masksToBounds = newValue > 0
        }
    }
    
    /**
     Add a border color functionality.
     */
    @IBInspectable var borderColor: UIColor {
        get {
            return UIColor(cgColor: layer.borderColor!)
        }
        set {
            layer.borderColor = newValue.cgColor
        }
    }
}


// ===========================================================
// MARK: - UIView
// ===========================================================

extension UIView{
    
    /**
     Add a border corner radius functionality.
     */
    @IBInspectable var cornerRadius: CGFloat {
        get {
            return layer.cornerRadius
        }
        set {
            layer.cornerRadius = newValue
            layer.masksToBounds = newValue > 0
        }
    }
    
    /**
     Hides the keyboard when user tap around on the screen.
     */
    func hideKeyboardWhenTappedAround() {
        let tap: UITapGestureRecognizer = UITapGestureRecognizer(target: self, action: #selector(UIView.dismissKeyboard))
        self.addGestureRecognizer(tap)
    }
    
    /**
     Hides the keyboard.
     */
    @objc func dismissKeyboard() {
        self.endEditing(true)
    }
    
    /**
     Create a solid status bar.
     
     - parameter backgroundColor: Status bar background color
     */
    func solidStatusBar(backgroundColor: UIColor) {
        // Create a solid color background for the status bar
        
        let statusFrame = CGRect(x: 0, y: 0, width: self.bounds.size.width, height: UIApplication.shared.statusBarFrame.size.height)
        
        let statusBar = UIView(frame: statusFrame)
        
        statusBar.backgroundColor = backgroundColor
        
        self.addSubview(statusBar)
    }
}


// ===========================================================
// MARK: - UIViewController
// ===========================================================

extension UIViewController {
    
    struct SizeClass {
        var width = UIUserInterfaceSizeClass.regular    // Horitzontal
        var height = UIUserInterfaceSizeClass.regular   // Vertical
    }
    
    /**
     Get UIView size class.
     
     - return: Struct sizeClass with width (horitzontal) and height (vertical)
     */
    func getSizeClass() -> SizeClass {
        return SizeClass(width: self.traitCollection.horizontalSizeClass, height: self.traitCollection.verticalSizeClass)
    }
    
    /**
     Lifting the view up.
     
     - parameter up:            True to move up, false to move down.
     - parameter moveValue:     Pixels movement.
     */
    func animateViewMoving (_ up:Bool, moveValue :CGFloat){
        let movementDuration:TimeInterval = 0.3
        let movement:CGFloat = ( up ? -moveValue : moveValue)
        UIView.beginAnimations( "animateView", context: nil)
        UIView.setAnimationBeginsFromCurrentState(true)
        UIView.setAnimationDuration(movementDuration )
        self.view.frame = self.view.frame.offsetBy(dx: 0,  dy: movement)
        UIView.commitAnimations()
    }
    
    /**
     Show alert like a pop up
     
     - parameter title:     Pop up title.
     - parameter message:   Pop up message.
     */
    func showPopupAlert(title: String?, message: String?, image: UIImage?, buttonLabel: String, buttonColor: UIColor?, style: UIAlertActionStyle, handler: PopupDialogButton.PopupDialogButtonAction?) {
        
        // Create the dialog
        let popup = PopupDialog(title: title, message: message, image: image, buttonAlignment: .horizontal, transitionStyle: .zoomIn, gestureDismissal: true)
        
        // Create first button
        if style == .cancel {
            let buttonOne = CancelButton(title: NSLocalizedString("CANCEL", comment: "Cancel an action"), action: nil)
            popup.addButton(buttonOne)
        }
        
        // Create second button
        let buttonTwo = DefaultButton(title: buttonLabel, action: handler)
        buttonTwo.titleColor = UIColor.primary.light
        popup.addButton(buttonTwo)
        
        // Present dialog
        if presentedViewController == nil {
            self.present(popup, animated: true, completion: nil)
        }
    }
    
    /**
     Show alert like a notification
     
     - parameter title:     Notification title.
     - parameter message:   Notification message.
     - parameter theme:     Theme.
     */
    func showNotificationAlert(title: String?, message: String?, theme: Theme) {
        
        let notification = MessageView.viewFromNib(layout: .tabView)
        notification.configureTheme(theme)
        notification.configureDropShadow()
        
        var iconText = ""
        if (theme == .info) {
            iconText = "🚀"
        } else {
            iconText = ["🤔", "😳", "🙄", "😶"].sm_random()!
        }
        notification.configureContent(title: title!, body: message!, iconText: iconText)
        notification.button?.isHidden = true
        
        var notificationConfig = SwiftMessages.defaultConfig
        
        if (theme == .info) {
            notificationConfig.presentationContext = .window(windowLevel: UIWindowLevelStatusBar)
        }
        
        SwiftMessages.show(config: notificationConfig, view: notification)
    }
}


// ===========================================================
// MARK: - URL
// ===========================================================

extension URL {
    
    /**
     Open URL available errors
     */
    enum OpenURLError: Error {
        case urlEmpty
        case urlInvalid
        case phoneInvalid
    }
    
    /**
     Initial configurations to open a website.
     
     - return: Error.
     */
    func openBrowser() -> Error? {
        
        if UIApplication.shared.canOpenURL(self) {
            UIApplication.shared.open(self, options: [:], completionHandler: nil)
            return nil
        } else if self.absoluteString.isEmpty {
            return OpenURLError.urlEmpty
        } else {
            return OpenURLError.urlInvalid
        }
    }
    
    /**
     Initial configurations to call a phone number. It's function fix
     the issue to exist various phone number formats.
     
     - parameters number: Original value, depends of country.
     - return: True if it is possible to use that format.
     */
    static func callPhone(number: String) -> Error? {
        
        var editedPhoneNumber = ""
        
        if number != "" {
            
            for i in number {
                
                switch (i){
                case "0","1","2","3","4","5","6","7","8","9" : editedPhoneNumber = editedPhoneNumber + String(i)
                case "+" : editedPhoneNumber = editedPhoneNumber + "00"
                default : print("Removed invalid character.")
                }
            }
            
            let phone = "tel://" + editedPhoneNumber
            let url = URL(string: phone)
            if let url = url {
                UIApplication.shared.open(url, options: [:], completionHandler: nil)
            } else {
                return OpenURLError.urlInvalid
            }
        } else {
            return OpenURLError.phoneInvalid
        }
        return nil
    }
    
}
