/**
 FirebaseHelper.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import Foundation
import Firebase
import GeoFire


class FirebaseHelper {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    static let sharedInstance = FirebaseHelper()
    
    let BASE_REF = Database.database().reference()
    let EVENT_REF = Database.database().reference().child("event")
    let GEOLOCATION_REF = Database.database().reference().child("geofire")
    let ALLEDITION_REF = Database.database().reference().child("alledition")
    
    private var geoFire: GeoFire
    
    // Incremental distance (for loop query each X km)
    private let incrementalQueryDistance = 50.0
    // Minimum distance (km)
    private let minQueryDistance = 10.0
    // Maximum distance (km)
    private let maxQueryDistance = 2000.0
    
    // Elements to get each query
    private let PAGESIZE = 30
    
    // Query number to control 2 or more query at the same time
    private var currentQueryNumber = 0
    
    // ===========================================================
    // MARK: - Constructors
    // ===========================================================
    
    /**
     Init geofire reference.
     */
    private init ()
    {
        geoFire = GeoFire(firebaseRef: GEOLOCATION_REF)
    }
    
    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    /**
     Sign in.
     */
    static func signIn(with credential: AuthCredential, completion: @escaping (_ error: Error?) -> Void ) {
        let anonymousUser = Auth.auth().currentUser
        
        Auth.auth().signIn(with: credential) { user, error in
            
            if let error = error {
                print(error.localizedDescription)
            } else {
                anonymousUser?.delete { error in
                    if let error = error {
                        print("Delete anonymous user: "+error.localizedDescription)
                    }
                }
            }
            completion(error)
        }
    }
    
    /**
     Get all events arround current location.
     
     - parameter center: Current location
     */
    func getAllEvents (center: CLLocation, completion: @escaping (_ event: Event) -> Void ) {
        UIApplication.shared.isNetworkActivityIndicatorVisible = true
        
        var keys = [String]()
        
        currentQueryNumber += 1
        let queryNumber = currentQueryNumber
        
        let circleQuery = geoFire.query(at: center, withRadius: minQueryDistance)
        
        circleQuery!.observe(.keyEntered, with: { key,location in
            keys.append(key!)
            self.EVENT_REF.child(key!).observeSingleEvent(of: .value, with: { snap in
                
                if let postDictionary = snap.value as? Dictionary<String, AnyObject> {
                    let key = snap.key
                    let event = Event(key: key, json: postDictionary)
                    
                    completion(event)
                }
            })
        })
        
        var queryDistance = minQueryDistance
        
        // Do something with list of keys.
        circleQuery!.observeReady({
            if queryNumber == self.currentQueryNumber {
                
                if keys.count > self.PAGESIZE || queryDistance > self.maxQueryDistance {
                    circleQuery?.removeAllObservers()
                    UIApplication.shared.isNetworkActivityIndicatorVisible = false
                } else {
                    queryDistance += self.incrementalQueryDistance
                    circleQuery?.radius = queryDistance
                }
            }
        })
    }
    
    /**
     Create new event and save to Firebase.
     
     - parameter event: New event (properties dictionary)
     */
    func createNewEvent(event: Dictionary<String, AnyObject>) {
        
        // Save the Event
        // EVENT_REF is the parent of the new Event: "events".
        // childByAutoId() saves the event and gives it its own ID.
        
        let firebaseNewEvent = EVENT_REF.childByAutoId()
        
        // setValue() saves to Firebase.
        
        firebaseNewEvent.setValue(event)
    }
}

