/**
 Address.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import Foundation
import CoreLocation


struct AddressKey {
    static let coordinates = "coordinates"
    static let thoroughfare = "thoroughfare"
    static let locality = "locality"
    static let subAdministrativeArea = "subAdministrativeArea"
    static let administrativeArea = "administrativeArea"
    static let postalCode = "postalCode"
    static let country = "country"
    static let isoCountryCode = "isoCountryCode"
}


enum AddressStringFormat {
    case full
    case large
    case medium
    case short
}


class Address {
    
    // ===========================================================
    // MARK: - Variables and constants
    // ===========================================================
    
    var location: CLLocation = CLLocation()
    var thoroughfare: String = ""
    var locality: String = ""
    var administrativeArea: String = ""
    var subAdministrativeArea: String = ""
    var postalCode: String = ""
    var country: String = ""
    var isoCountryCode: String = ""
    
    // ===========================================================
    // MARK: - Constructors
    // ===========================================================
    
    /**
     Init with JSON dictionary.
     */
    init(json: Dictionary<String, AnyObject>) {
        for (elementKey, element) in json {
            switch elementKey {
            case AddressKey.coordinates:
                let coordinates = element as! [Double]
                self.location = CLLocation(latitude: coordinates[0], longitude: coordinates[1])
            case AddressKey.thoroughfare:
                self.thoroughfare = element as! String
            case AddressKey.locality:
                self.locality = element as! String
            case AddressKey.subAdministrativeArea:
                self.subAdministrativeArea = element as! String
            case AddressKey.administrativeArea:
                self.administrativeArea = element as! String
            case AddressKey.postalCode:
                self.postalCode = element as! String
            case AddressKey.country:
                self.country = element as! String
            case AddressKey.isoCountryCode:
                self.isoCountryCode = element as! String
            default:
                print("No valid JSON element")
            }
        }
    }
    
    /**
     Init with location inverse.
     */
    init(location: CLLocation, completion: @escaping (_ error: Error?) -> Void) {
        
        CLGeocoder().reverseGeocodeLocation(location, completionHandler: { placemarks, error in
            
            if let error = error {
                print(error.localizedDescription)
                completion(error)
                return
            }
            
            if let placeMark = placemarks?[0] {
                self.uploadAdress(from: placeMark)
            }
            
            completion(error)
        })
    }
    
    /**
     Init with location forward.
     */
    init(address: String, completion: @escaping (_ error: Error?) -> Void) {
        
        CLGeocoder().geocodeAddressString(address, completionHandler: { (placemarks, error) in
            
            if let error = error {
                print(error.localizedDescription)
                completion(error)
                return
            }
            
            if let placeMark = placemarks?[0] {
                self.uploadAdress(from: placeMark)
            }
            
            completion(error)
        })
    }
    
    // ===========================================================
    // MARK: - Methods
    // ===========================================================
    
    /**
     Upload an address with placemark data.
     
     - parameter placeMark: Placemark with all new data.
    */
    private func uploadAdress(from placeMark: CLPlacemark) {
        // Location address
        if let location = placeMark.location {
            print("\nlat: \(location.coordinate.latitude), long: \(location.coordinate.longitude)")
            self.location = location
        }
        
        // Street address
        if let thoroughfare = placeMark.thoroughfare {
            print(thoroughfare)
            self.thoroughfare = thoroughfare
        }
        
        // Locality
        if let locality = placeMark.locality {
            print(locality)
            self.locality = locality
        }
        
        // Subadministrative Area (Provincia)
        if let subAdministrativeArea = placeMark.subAdministrativeArea {
            print(subAdministrativeArea)
            self.subAdministrativeArea = subAdministrativeArea
        }
        
        // Postal code
        if let postalCode = placeMark.postalCode {
            print(postalCode)
            self.postalCode = postalCode
        }
        
        // Administrative Area (Comunitat autonoma)
        if let administrativeArea = placeMark.administrativeArea {
            print(administrativeArea)
            self.administrativeArea = administrativeArea
        }
        
        // Country
        if let country = placeMark.country {
            print(country)
            self.country = country
        }
        
        // Country code
        if let isoCountryCode = placeMark.isoCountryCode {
            print(isoCountryCode)
            self.isoCountryCode = isoCountryCode
        }
    }
    
    /**
     Create differents string outputs of address.
     
     - parameter format: Address format.
     - return: String formatted.
     */
    func toString(format: AddressStringFormat) -> String {
        switch format {
        case .full:
            return String(format: "%@, %@, %@, %@, %@", thoroughfare, locality, administrativeArea, postalCode, country)
        case .large:
            return String(format: "%@, %@, %@, %@", thoroughfare, locality, administrativeArea, postalCode)
        case .medium:
            return String(format: "%@, %@, %@", locality, administrativeArea, postalCode)
        case .short:
            return String(format: "%@, %@", locality, country)
        }
    }
}
