/**
 Regex.swift
 
 This file is part of the Eventhon package.
 @copyright (c) 2017-present, Eventhon. All rights reserved.
 @author Raul Romero Rostan <rromerorostan@gmail.com>
 @updated 2017-11-02
 
 For the full copyright and license information, please view the LICENSE
 file that was distributed with this source code.
 */

import Foundation


/**
 Regular expression basic utilities.
 */
class Regex {
  
  /**
   Check password complexity.
   
   - parameter password:         Password to test
   - parameter length:           Password min length
   - parameter patternsToEscape: Patterns that password must not contains
   - parameter caseSensitivty:   Specify if password must conforms case sensitivity or not
   - parameter numericDigits:    Specify if password must conforms contains numeric digits or not
   
   - returns: Boolean that describes if password is valid or not
   */
  class func checkPasswordComplexity(password: String, length: Int, patternsToEscape: [String], caseSensitivty: Bool, numericDigits: Bool) -> Bool {
    if (password.count < length) {
      return false
    }
    if caseSensitivty {
      let hasUpperCase = Regex.matchesForRegexInText(regex: "[A-Z]", text: password).count > 0
      if !hasUpperCase {
        return false
      }
      let hasLowerCase = Regex.matchesForRegexInText(regex: "[a-z]", text: password).count > 0
      if !hasLowerCase {
        return false
      }
    }
    if numericDigits {
      let hasNumbers = Regex.matchesForRegexInText(regex: "\\d", text: password).count > 0
      if !hasNumbers {
        return false
      }
    }
    if patternsToEscape.count > 0 {
      let passwordLowerCase = password.lowercased()
      for pattern in patternsToEscape {
        let hasMatchesWithPattern = Regex.matchesForRegexInText(regex: pattern, text: passwordLowerCase).count > 0
        if hasMatchesWithPattern {
          return false
        }
      }
    }
    return true
  }
  
  /**
   Matches of regular expression in text..
   
   - parameter regex:     Regular expression.
   - parameter text:      Text to check.
   
   - returns: String arrray with substrings that have passed the test.
   */
  class func matchesForRegexInText(regex: String, text: String) -> [String] {
    do {
      let regex = try NSRegularExpression(pattern: regex, options: [])
      let nsString = text as NSString
      let results = regex.matches(in: text,
                                  options: [], range: NSMakeRange(0, nsString.length))
      return results.map { nsString.substring(with: $0.range)}
    } catch let error as NSError {
      print("invalid regex: \(error.localizedDescription)")
      return []
    }
  }
  
  /**
   Check regular expression.
   
   - parameter regex:     Regular expression.
   - parameter text:      Text to check.
   
   - returns: Boolean if text has passed the regular expression.
   */
  class func checkRegexInText(regex: String, text: String) -> Bool {
    do {
      let regex = try NSRegularExpression(pattern: regex, options: .caseInsensitive)
      return regex.firstMatch(in: text, options: NSRegularExpression.MatchingOptions(rawValue: 0), range: NSMakeRange(0, text.count)) != nil
    } catch {
      return false
    }
  }
}
